from typing import Any


def raise_has_not_attr(obj: Any, name: str) -> bool:
    if not hasattr(obj, name):
        if hasattr(obj, '__class__'):
            obj_name = obj.__class__.__name__
        elif hasattr(obj, '__name__'):
            obj_name = obj.__name__
        else:
            obj_name = str(obj)
        raise AttributeError(f"'{obj_name}' has no attribute '{name}'!")
    return True



import inspect
import logging
from ipaddress import IPv4Address
from pathlib import Path
from typing import Any, Optional, Union, Sequence

from sqlalchemy import create_engine, Table
from sqlalchemy.orm import sessionmaker, declarative_base, DeclarativeMeta as _DeclarativeMeta, Session
from sqlalchemy.ext.declarative import declarative_base

from wiederverwendbar.sqlalchemy.settings import SqlalchemySettings

logger = logging.getLogger(__name__)


class DeclarativeMeta(_DeclarativeMeta):
    def __init__(cls, classname: Any, bases: Any, dict_: Any, **kw: Any) -> None:
        db = None
        for base in bases:
            if hasattr(base, "db"):
                db = base.db
        if db is None:
            stack = inspect.stack()
            for frame in stack:
                if frame.function == "__init__":
                    db = frame.frame.f_locals.get("self", None)
                    if isinstance(db, SqlalchemyDb):
                        break
        super().__init__(classname=classname, bases=bases, dict_=dict_, **kw)

        cls.db: SqlalchemyDb = db


class SqlalchemyDb:
    def __init__(self,
                 file: Optional[Path] = None,
                 host: Union[None, IPv4Address, str] = None,
                 port: Optional[int] = None,
                 protocol: Optional[str] = None,
                 name: Optional[str] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 echo: Optional[bool] = None,
                 settings: Optional[SqlalchemySettings] = None):
        """
        Create a new Sqlalchemy Database

        :param host: Host to connect to database
        :param port: Port to connect to database
        :param protocol: Protocol to connect to database
        :param name: Name of the database
        :param username: User to connect to database
        :param password: Password to connect to database
        :param echo: Echo SQL queries to console
        :param settings: Sqlalchemy Settings
        """

        self.settings: SqlalchemySettings = settings or SqlalchemySettings()
        self.file: Optional[Path] = file or self.settings.db_file
        self.host: Union[IPv4Address, str, None] = host or self.settings.db_host
        self.port: Optional[int] = port or self.settings.db_port
        self.protocol: Optional[str] = protocol or self.settings.db_protocol
        self.name: Optional[str] = name or self.settings.db_name
        self.username: Optional[str] = username or self.settings.db_username
        self.password: Optional[str] = password or self.settings.db_password
        self.echo: bool = echo or self.settings.db_echo

        logger.debug(f"Create {self}")

        self.engine = create_engine(self.connection_string, echo=self.echo)
        self.session_maker = sessionmaker(bind=self.engine)
        self.Base: DeclarativeMeta = declarative_base(metaclass=DeclarativeMeta)
        self.session_maker.configure(binds={self.Base: self.engine})

    def __str__(self):
        return f"{self.__class__.__name__}({self.connection_string_printable})"

    def get_connection_string(self, printable: bool = False) -> str:
        """
        Get the Connection String

        :param printable: If True, the password will be hidden
        :return: str
        """

        connection_string = f"{self.protocol}://"
        if self.protocol == "sqlite":
            if self.file is not None:
                connection_string += f"/{self.file}"
        else:
            if self.username is not None:
                connection_string += f"{self.username}"
            if self.password is not None:
                connection_string += ":"
                if printable:
                    connection_string += "***"
                else:
                    connection_string += self.password
            if self.host is None:
                raise RuntimeError(f"No host specified for {self}")
            connection_string += f"@{self.host}"
            if self.port is None:
                raise RuntimeError(f"No port specified for {self}")
            connection_string += f":{self.port}"
            if self.name is None:
                raise RuntimeError(f"No name specified for {self}")
            connection_string += f"/{self.name}"
        return connection_string

    @property
    def connection_string(self) -> str:
        """
        Get the Connection String

        :return: str
        """

        return self.get_connection_string()

    @property
    def connection_string_printable(self) -> str:
        """
        Get the Connection String with Password hidden

        :return: str
        """

        return self.get_connection_string(printable=True)

    def create_all(self,
                   tables: Optional[Sequence[Table]] = None,
                   check_first: bool = True) -> None:
        """
        Create all Tables

        :param tables: List of Tables to create. If None, all Tables will be created.
        :param check_first: Check if Tables exist before creating them.
        :return: None
        """

        logger.debug(f"Create all for {self}")
        self.Base.metadata.create_all(bind=self.engine, tables=tables, checkfirst=check_first)

    def session(self) -> Session:
        """
        Create a new Session

        :return: Session
        """

        logger.debug(f"Create Session for {self}")
        return self.session_maker()



import logging
from pathlib import Path
from typing import Optional, Union
from ipaddress import IPv4Address

from pymongo import MongoClient
from pymongo.database import Database
from pymongo.errors import PyMongoError
from mongoengine import DEFAULT_CONNECTION_NAME, connect, disconnect

from wiederverwendbar.mongoengine.backup import dump, restore
from wiederverwendbar.mongoengine.settings import MongoengineSettings

logger = logging.getLogger(__name__)


class MongoengineDb:
    def __init__(self,
                 name: Optional[str] = None,
                 host: Union[None, IPv4Address, str] = None,
                 port: Optional[int] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 auth_source: Optional[str] = None,
                 timeout: Optional[int] = None,
                 test: Optional[bool] = None,
                 auto_connect: Optional[bool] = None,
                 settings: Optional[MongoengineSettings] = None):
        """
        Create a new Mongoengine Database

        :param name: Database Name(aka Alias in Mongoengine)
        :param settings: Mongoengine Settings
        """

        self.name: str = name or DEFAULT_CONNECTION_NAME
        self.settings: MongoengineSettings = settings or MongoengineSettings()
        self.host: Union[IPv4Address, str] = host or self.settings.db_host
        self.port: int = port or self.settings.db_port
        self.db_name: str = name or self.settings.db_name
        self.username: str = username or self.settings.db_username
        self.password: str = password or self.settings.db_password
        self.auth_source: str = auth_source or self.settings.db_auth_source
        self.timeout: int = timeout or self.settings.db_timeout
        self.db_test: bool = test or self.settings.db_test
        self.auto_connect: bool = auto_connect or self.settings.db_auto_connect

        logger.debug(f"Create {self}")

        # connect to database
        self._client: Optional[MongoClient] = None
        self._db: Optional[Database] = None

        if self.auto_connect:
            self.connect()

    def __str__(self):
        return f"{self.__class__.__name__}(name={self.name}, host={self.host}, port={self.port})"

    def __del__(self):
        if self.is_connected:
            self.disconnect()

    @property
    def connection_string(self) -> str:
        """
        Get the Connection String

        :return: str
        """

        connection_string = "mongodb://"
        if self.username and self.password:
            connection_string += f"{self.username}:{self.password}@"
        connection_string += f"{self.host}:{self.port}/{self.db_name}?authSource={self.auth_source}"

        return connection_string

    @property
    def client(self) -> MongoClient:
        """
        Get the Database Client

        :return: MongoClient
        """

        if self._client is None:
            self.connect()
        return self._client

    @property
    def db(self) -> Database:
        """
        Get the Database

        :return: Database
        """

        if self._db is None:
            self.connect()
        return self._db

    @property
    def is_connected(self) -> bool:
        """
        Check if the database is connected

        :return: bool
        """

        if self._client is None:
            return False
        if self._db is None:
            return False
        return True

    def connect(self) -> None:
        """
        Connect to the database

        :return: None
        """

        logger.debug(f"Connect to {self} ...")

        if self.is_connected:
            raise RuntimeError(f"Already connected to {self}")

        self._client = connect(db=self.db_name,
                               alias=self.name,
                               host=self.host,
                               port=self.port,
                               username=self.username,
                               password=self.password,
                               authSource=self.auth_source,
                               serverSelectionTimeoutMS=self.timeout)
        self._db = self.client[self.db_name]

        if self.db_test:
            self.test()

    def test(self):
        """
        Test the database connection

        :return: None
        """

        logger.debug(f"Testing {self} ...")

        if not self.is_connected:
            raise RuntimeError(f"Not connected to {self}")

        try:
            self.db.list_collection_names()
        except PyMongoError as e:
            raise RuntimeError(f"Could not connect to database: {e}")

    def disconnect(self) -> None:
        """
        Disconnect from the database

        :return: None
        """

        logger.debug(f"Disconnect from {self} ...")

        if not self.is_connected:
            raise RuntimeError(f"Not connected to {self}")

        disconnect(alias=self.name)
        self._client = None
        self._db = None

    def dump(self: Database, path: Union[str, Path], overwrite: bool = False) -> None:
        """
        MongoDB Dump
        :param path: Database dump path
        :param overwrite: Overwrite existing files
        :return: None
        """

        dump(db=self.db, path=path, overwrite=overwrite)

    def restore(self: Database, path: Union[str, Path], overwrite: bool = False) -> None:
        """
        MongoDB Restore
        :param path: Database dump path
        :param overwrite: Overwrite existing collections
        :return: None
        """

        restore(db=self.db, path=path, overwrite=overwrite)



import logging

from mongoengine import StringField

from wiederverwendbar.mongoengine import MongoengineDbSingleton, MongoengineLogDocument, MongoengineLogHandler

MongoengineDbSingleton(init=True)


class MyMongoengineLogDocument(MongoengineLogDocument):
    meta = {"collection": "log.test"}

    test = StringField(required=True)


logger = logging.getLogger("test123")

logger.addHandler(logging.StreamHandler())
logger.addHandler(MongoengineLogHandler(document=MyMongoengineLogDocument, document_kwargs={"test": "test"}))

logger.setLevel(logging.DEBUG)

if __name__ == '__main__':
    logger.debug("Hello")
    logger.debug("Hello")



from typing import Any, Optional, Callable, Union, TYPE_CHECKING

from sqlalchemy import inspect
from sqlalchemy.sql import ColumnExpressionArgument
from sqlalchemy.orm import Session, QueryableAttribute, InstrumentedAttribute
from sqlalchemy.orm.exc import DetachedInstanceError

from wiederverwendbar.default import Default
from wiederverwendbar.sqlalchemy.raise_has_not_attr import raise_has_not_attr

if TYPE_CHECKING:
    from wiederverwendbar.sqlalchemy.db import SqlalchemyDb


class SelectionError(Exception):
    """Failed to select column value."""


class SerializationError(Exception):
    """Failed to serialize column value."""


class DeserializationError(Exception):
    """Failed to serialize column value."""


class MappingError(Exception):
    """Failed to map column value."""


class DictColumn:
    def __init__(self,
                 key: str,
                 is_primary_key: Union[bool, Default] = Default(),
                 selector: Union[Callable[..., Any], None, Default] = Default(),
                 serializer: Union[Callable[[...], Any], None, Default] = Default(),
                 deserializer: Union[Callable[[...], Any], None, Default] = Default(),
                 mapper: Union[Callable[..., Any], None, Default] = Default()) -> None:
        """
        DictColumn
        Used to select and serialize attributes from the ORM.

        :param key: Key in serial_obj.
        :param is_primary_key: Indicate if column is primary key.
        :param selector: Python code to select value from ORM.
        :param serializer: Python code to serialize value after selecting from ORM.
        :param deserializer: Python code to deserialize value for mapping.
        :param mapper: Python code to map value to ORM after deserializing.
        """

        # key
        self.key = key

        # is_primary_key
        if type(is_primary_key) is Default:
            is_primary_key = False
        self.is_primary_key = is_primary_key

        # selector
        if type(selector) is Default:
            selector = lambda **kw: getattr(kw["base"], kw["key"])
        self.selector = selector

        # serializer
        if type(serializer) is Default:
            serializer = None
        self.serializer = serializer

        # deserializer
        if type(deserializer) is Default:
            deserializer = None
        self.deserializer = deserializer

        # mapper
        if type(mapper) is Default:
            mapper = lambda **kw: raise_has_not_attr(kw["base"], kw["key"]) and setattr(kw["base"], kw["key"], kw["value"])
        self.mapper = mapper

    def __str__(self):
        s = f"{self.__class__.__name__}(key={self.key}"
        s += f", selector={self.selector is not None}"
        s += f", serializer={self.serializer is not None}"
        s += f", deserializer={self.deserializer is not None}"
        s += f", mapper={self.mapper is not None}"
        s += ")"
        return s

    def __eq__(self, other) -> bool:
        if type(other) is not DictColumn:
            return self.key == other
        else:
            return self.key == other.key and self.selector == other.selector and self.serializer == other.serializer

    def select(self, base: "Base") -> Any:
        if not isinstance(base, Base):
            raise AttributeError(f"Attribute 'base' must be a instance of Base, not {type(base)}")
        if self.selector is None:
            value = ...
        else:
            try:
                value = self.selector(key=self.key, base=base)
            except Exception as e:
                raise SelectionError(e) from e
        return value

    def serialize(self, value: Any) -> dict[str, Any]:
        # serialize
        try:
            if self.serializer is not None:
                value = self.serializer(value)
        except Exception as e:
            raise SerializationError(e) from e
        return value

    def deserialize(self, value: Any) -> Any:
        # deserialize
        try:
            if self.deserializer is not None:
                value = self.deserializer(value)
        except Exception as e:
            raise DeserializationError(e) from e
        return value

    def map(self, value: Any, base: "Base") -> None:
        if not isinstance(base, Base):
            raise AttributeError(f"Attribute 'base' must be a instance of Base, not {type(base)}")
        if self.mapper is not None:
            try:
                self.mapper(key=self.key, value=value, base=base)
            except Exception as e:
                raise SelectionError(e) from e


class Base:
    __abstract__ = True
    __allow_unmapped__ = True
    __str_columns__: list[str] = []
    db: "SqlalchemyDb"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}({', '.join([f'{column_name}={getattr(self, column_name)}' for column_name in self.__str_columns__])})"

    def __repr__(self):
        return self.__str__()

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except DetachedInstanceError:
            with self.db.session() as session:
                my_id = getattr(self, "id", None)
                if my_id is None:
                    raise Exception(f"{self.__class__.__name__} has no id!")
                obj = session.query(self.__class__).filter_by(id=my_id).first()
                if obj is None:
                    raise Exception(f"{self.__class__.__name__} with id {my_id} not found!")
                return getattr(obj, item)

    @classmethod
    def _default_dict_columns(cls) -> list[DictColumn]:
        # get inspection obj
        cls_inspect = inspect(cls)

        # create dict columns
        dict_columns = []
        for column_attrs in cls_inspect.mapper.column_attrs:
            dict_column = DictColumn(key=column_attrs.key, is_primary_key=column_attrs.expression.primary_key)
            dict_columns.append(dict_column)

        return dict_columns

    def as_dict(self, dict_columns: Union[list[DictColumn], Default] = Default()) -> dict[str, Any]:
        # create DictColumns if not given
        if type(dict_columns) is Default:
            dict_columns = self._default_dict_columns()

        # progress
        serial_obj = {}
        for dict_column in dict_columns:
            # check if key already in serial_obj
            if dict_column.key in serial_obj:
                raise SerializationError(f"Key '{self.key}' is already exist in serial_obj!")

            # select from base
            value = dict_column.select(base=self)
            if value is ...:
                continue

            # serialize to serial_obj
            serial_obj[dict_column.key] = dict_column.serialize(value=value)
        return serial_obj

    @classmethod
    def from_dict(cls, serial_obj: dict[str, Any], dict_columns: Union[list[DictColumn], Default] = Default()) -> "Base":
        # create DictColumns if not given
        if type(dict_columns) is Default:
            dict_columns = cls._default_dict_columns()

        # built get kwargs
        get_kwargs = {}
        for dict_column in dict_columns:
            if not dict_column.is_primary_key:
                continue

            # select from serial_obj
            value = serial_obj[dict_column.key]

            # deserialize
            get_kwargs[dict_column.key] = dict_column.deserialize(value=value)

        # get base
        if get_kwargs:
            base = cls.get(**get_kwargs)
        else:
            base = cls()

        # progress
        for dict_column in dict_columns:
            # select from serial_obj
            value = serial_obj[dict_column.key]

            # deserialize
            value = dict_column.deserialize(value=value)

            # map
            dict_column.map(value=value, base=base)
        return base

    @classmethod
    def session(cls, session: Union[Session] = None) -> tuple[bool, Session]:
        # create new session
        session_created = False
        if session is None:
            session = cls.db.session()
            session_created = True
        return session_created, session

    @classmethod
    def session_close(cls, session_created: bool, session: Session) -> None:
        # close session
        if session_created:
            session.close()

    @classmethod
    def new(cls,
            session: Optional[Session] = None,
            **kwargs) -> "Base":
        session_created, session = cls.session(session=session)

        # noinspection PyArgumentList
        obj: Base = cls(**kwargs)
        session.add(obj)
        session.commit()
        session.refresh(obj)

        cls.session_close(session_created=session_created, session=session)

        return obj

    @classmethod
    def length(cls,
               *criterion: Union[ColumnExpressionArgument[bool], bool],
               session: Optional[Session] = None,
               **kwargs: Any) -> int:
        session_created, session = cls.session(session=session)

        if criterion:
            length = session.query(cls).filter(*criterion, **kwargs).count()
        else:
            length = session.query(cls).filter_by(**kwargs).count()

        cls.session_close(session_created=session_created, session=session)

        return length

    @classmethod
    def get_all(cls,
                *criterion: Union[ColumnExpressionArgument[bool], bool],
                order_by: Union[str, QueryableAttribute, None] = None,
                order_desc: bool = False,
                rows_per_page: Optional[int] = None,
                page: int = 1,
                fill_empty: bool = False,
                as_dict: bool = False,
                dict_columns: Union[list[DictColumn], Default] = Default(),
                session: Optional[Session] = None,
                **kwargs: Any) -> list[Union["Base", None, dict[str, Any]]]:
        session_created, session = cls.session(session=session)

        # get order by
        if order_by is not None:
            if type(order_by) is str:
                order_by = getattr(cls, order_by)
            if not isinstance(order_by, QueryableAttribute):
                raise AttributeError(f"Attribute 'order_by' must be a instance of QueryableAttribute, not {type(order_by)}")
            if order_desc:
                order_by = order_by.desc()

        if rows_per_page is not None:
            # get length
            length = cls.length(*criterion, **kwargs)
            lst = []
            if length == 0:
                return lst
            offset = (page - 1) * rows_per_page

            # add empty rows before partial select
            if fill_empty:
                # noinspection PyTypeChecker
                for i in range(offset):
                    if as_dict:
                        lst.append({})
                    else:
                        lst.append(None)

            # get rows
            if criterion:
                lst_partial: list[Any] = session.query(cls).filter(*criterion, **kwargs).order_by(order_by).limit(rows_per_page).offset(offset).all()
            else:
                lst_partial: list[Any] = session.query(cls).filter_by(**kwargs).order_by(order_by).limit(rows_per_page).offset(offset).all()

            for row in lst_partial:
                lst.append(row.as_dict(dict_columns=dict_columns) if as_dict else row)

            # add empty rows after partial select
            if fill_empty:
                # noinspection PyTypeChecker
                for i in range(offset + rows_per_page, length):
                    if as_dict:
                        lst.append({})
                    else:
                        lst.append(None)
        else:
            if criterion:
                lst: list[Any] = session.query(cls).order_by(order_by).filter(*criterion, **kwargs).all()
            else:
                lst: list[Any] = session.query(cls).order_by(order_by).filter_by(**kwargs).all()
            if as_dict:
                lst = [row.as_dict(dict_columns=dict_columns) for row in lst]

        cls.session_close(session_created=session_created, session=session)

        return lst

    @classmethod
    def get(cls,
            *criterion: Union[ColumnExpressionArgument[bool], bool],
            as_dict: bool = False,
            dict_columns: Union[list[DictColumn], Default] = Default(),
            session: Optional[Session] = None,
            **kwargs: Any) -> Union["Base", None, dict[str, Any]]:
        session_created, session = cls.session(session=session)

        if criterion:
            obj: Any = session.query(cls).filter(*criterion, **kwargs).first()
        else:
            obj: Any = session.query(cls).filter_by(**kwargs).first()

        if as_dict:
            if obj is None:
                obj = None
            else:
                obj = obj.as_dict(dict_columns=dict_columns)

        cls.session_close(session_created=session_created, session=session)

        return obj

    def save(self, session: Optional[Session] = None) -> "Base":
        session_created, session = self.session(session=session)

        session.add(self)
        session.commit()
        session.refresh(self)

        self.session_close(session_created=session_created, session=session)

        return self

    def update(self, session: Optional[Session] = None, **kwargs) -> "Base":
        session_created, session = self.session(session=session)

        for column, value in kwargs.items():
            setattr(self, column, value)

        self.save()

        self.session_close(session_created=session_created, session=session)

        return self

    def delete(self, session: Optional[Session] = None) -> None:
        session_created, session = self.session(session=session)

        session.delete(self)
        session.commit()

        self.session_close(session_created=session_created, session=session)



from typing import Any

import starlette_admin as sa
from starlette_admin.contrib.mongoengine.converters import ModelConverter, converts
from mongoengine import BooleanField as MongoengineBooleanField

from wiederverwendbar.mongoengine.fields.boolean_also_field import BooleanAlsoField as MongoengineBooleanAlsoField
from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field.field import BooleanAlsoField as StarletteAdminBooleanAlsoField


class BooleanAlsoConverter(ModelConverter):
    @converts(MongoengineBooleanField, MongoengineBooleanAlsoField)
    def conv_boolean_field(self, *args: Any, **kwargs: Any) -> sa.BaseField:
        field = kwargs['field']
        also=""
        if isinstance(field, MongoengineBooleanAlsoField):
            also=field.also.name
        return StarletteAdminBooleanAlsoField(**self._field_common(*args, **kwargs), also=also)



import logging
import subprocess

logger = logging.getLogger(__name__)


def run_command(cmd: list[str], expected_exit_code=0) -> tuple[bool, list[str], list[str]]:
    """
    Run a command and log its output.

    :param cmd: Command as list of strings
    :param expected_exit_code: Expected exit code of command
    :return: True if command was successful, False otherwise
    """

    cmd_str = " ".join(cmd)
    logger.debug(f"Run command: {cmd_str}")

    # run command
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # read output from pip command and log it
    stdout = []
    stderr = []
    with process.stdout and process.stderr:
        for line in iter(process.stdout.readline, b''):
            line_str = line.decode("cp437").strip()
            logger.debug("stdout: " + line_str)
            stdout.append(line_str)

        for line in iter(process.stderr.readline, b''):
            line_str = line.decode("cp437").strip()
            logger.debug("stderr: " + line_str)
            stderr.append(line_str)

    exit_code = process.wait()

    # check exit status
    if exit_code != expected_exit_code:
        logger.error(f"Command '{cmd_str}' has exit status '{exit_code}' but expected exit status is '{expected_exit_code}'.")
        return False, stdout, stderr

    return True, stdout, stderr



import tempfile
from pathlib import Path
from typing import Literal

AVAILABLE_UNITS = Literal['B', 'KB', 'MB', 'GB']


def test_file(size: int = 1, unit: AVAILABLE_UNITS = 'MB') -> Path:
    """
    Generate a test file with the given size and unit.

    :param size: Size of the file
    :param unit: Unit of the size
    :return: Path to the generated file
    """

    if unit == 'B':
        file_size_in_bytes = size
    elif unit == 'KB':
        file_size_in_bytes = size * 1024
    elif unit == 'MB':
        file_size_in_bytes = size * 1024 * 1024
    elif unit == 'GB':
        file_size_in_bytes = size * 1024 * 1024 * 1024
    else:
        raise ValueError('Invalid unit')

    with tempfile.NamedTemporaryFile(delete=False) as tf:
        tf.seek(file_size_in_bytes - 1)
        tf.write(b'0')

        return Path(tf.name)



from wiederverwendbar.starlette_admin.action_log.logger import WebsocketHandler, ActionLoggerCommand, ActionLoggerResponse, ActionSubLogger, ActionSubLoggerContext, ActionLogger, StepCommand, NextStepCommand, \
    IncreaseStepsCommand, FinalizeCommand, FormCommand, ExitCommand, ConfirmCommand, DownloadCommand, YesNoCommand
from wiederverwendbar.starlette_admin.action_log.admin import ActionLogAdmin
from wiederverwendbar.starlette_admin.action_log.thread import ActionThread



from wiederverwendbar.post_init import PostInit

class MyParentClass(PostInit):
    def __init__(self, parent_attr: str):
        super().__init__()

        self.parent_attr = parent_attr

        print(f"Parent attribute: {parent_attr}")

    def __post_init__(self):
        print(f"Parent post init attribute: {self.parent_attr}")


class MyClass(MyParentClass):
    def __init__(self, my_attr: str, parent_attr: str):
        super().__init__(parent_attr=parent_attr)

        self.my_attr = my_attr

        print(f"My attribute: {my_attr}")

    def __post_init__(self):
        super().__post_init__()
        print(f"My post init attribute: {self.my_attr}")


if __name__ == '__main__':
    my_class = MyClass(my_attr="My attribute", parent_attr="Parent attribute")


import logging
import sys
from ipaddress import IPv4Network, IPv4Address, AddressValueError

from wiederverwendbar.functions.run_command import run_command
from wiederverwendbar.pydantic.indexable_model import IndexableModel
from wiederverwendbar.functions.admin import is_admin

logger = logging.getLogger(__name__)


class Route(IndexableModel):
    target: IPv4Network
    gateway: IPv4Address


def _win_list_routes() -> list[Route]:
    """
    List all IPv4 routes on Windows.

    :return: List of Route objects
    """

    logger.debug("List all IPv4 routes on Windows.")

    # run route print -4
    success, stdout, stderr = run_command(cmd=["route", "print", "-4"])

    if not success:
        raise SystemExit(1)

    # find route section in stdout
    raw_routes = []
    separator_line_count = 0
    skip_next_lines = 0
    caption_skipped = False
    for line in stdout:
        if skip_next_lines > 0:
            skip_next_lines -= 1
            continue
        if line.startswith("==="):
            separator_line_count += 1
            continue
        if separator_line_count == 3:
            if not caption_skipped:
                skip_next_lines = 1
                caption_skipped = True
                continue
            raw_routes.append(line)

    # parse routes
    routes: list[Route] = []
    for raw_route in raw_routes:
        route_list = raw_route.split()
        # parse cidr
        mask_str = route_list[1]
        mask_list = [int(x) for x in mask_str.split(".")]
        cidr = sum((bin(x).count('1') for x in mask_list))

        # parse target
        target_str = route_list[0] + "/" + str(cidr)
        target = IPv4Network(target_str)

        # parse gateway
        try:
            gateway = IPv4Address(route_list[2])
        except AddressValueError:
            # take interface address as gateway
            gateway = IPv4Address(route_list[-2])

        route = Route(target=target, gateway=gateway)
        routes.append(route)

    logger.debug(f"Found {len(routes)} routes.")
    return routes


def list_routes() -> list[Route]:
    if sys.platform == "win32":
        return _win_list_routes()
    else:
        raise NotImplementedError("This function is not implemented for this platform.")


def _win_create_route(route: Route) -> bool:
    """
    Create a route on Windows.

    :param route: Route object
    :return: True if route was created successfully, False otherwise
    """

    logger.debug(f"Create route: {route}")

    # list existing routes
    existing_routes = _win_list_routes()

    # check if route already exists
    found_existing_route = None
    for existing_route in existing_routes:
        if existing_route.target == route.target:
            found_existing_route = existing_route
            break

    if found_existing_route is not None:
        logger.warning(f"Route {route} already exists. Found existing route: {found_existing_route}")

        return False

    # parse cmd
    cmd = ["route", "add", str(route.target), "mask", str(route.target.netmask), str(route.gateway)]

    success, stdout, stderr = run_command(cmd=cmd)

    # check if 'OK!' is in stdout
    if "OK!" not in stdout:
        success = False

    if not success:
        logger.error(f"Failed to create route {route}.")
        return False

    logger.info(f"Route {route} created.")
    return True


def create_route(route: Route) -> bool:
    """
    Create a route.

    :param route: Route object
    :return: True if route was created successfully, False otherwise
    """

    if not is_admin():
        logger.error("This function requires admin privileges.")
        raise SystemExit(1)

    if sys.platform == "win32":
        return _win_create_route(route=route)
    else:
        raise NotImplementedError("This function is not implemented for this platform.")


def _win_delete_route(route: Route) -> bool:
    """
    Create a route on Windows.

    :param route: Route object
    :return: True if route was created successfully, False otherwise
    """

    logger.debug(f"Create route: {route}")

    # list existing routes
    existing_routes = _win_list_routes()

    # check if route not exists
    found_existing_route = None
    for existing_route in existing_routes:
        if existing_route.target == route.target:
            found_existing_route = existing_route
            break

    if found_existing_route is None:
        logger.warning(f"Route {route} does not exist.")
        return False

    # parse cmd
    cmd = ["route", "delete", str(route.target), str(route.gateway)]

    success, stdout, stderr = run_command(cmd=cmd)

    # check if 'OK!' is in stdout
    if "OK!" not in stdout:
        success = False

    if not success:
        logger.error(f"Failed to delete route {route}.")
        return False

    logger.info(f"Route {route} deleted.")

    return True


def delete_route(route: Route) -> bool:
    if not is_admin():
        logger.error("This function requires admin privileges.")
        raise SystemExit(1)

    if sys.platform == "win32":
        return _win_delete_route(route=route)
    else:
        raise NotImplementedError("This function is not implemented for this platform.")



import logging
from typing import Union


def logger_exists(logger_or_name: Union[str, logging.Logger]) -> bool:
    """
    Check if logger exists.

    :param logger_or_name: Logger or name of logger to check.
    :return: True if logger exists, False otherwise.
    """

    if isinstance(logger_or_name, logging.Logger):
        logger_name = logger_or_name.name
    elif isinstance(logger_or_name, str):
        logger_name = logger_or_name
    else:
        raise TypeError(f"Expected 'str' or 'logging.Logger', got '{type(logger_or_name)}'.")
    return logger_name in logging.root.manager.loggerDict


def remove_logger(*loggers_or_names: Union[str, logging.Logger]) -> None:
    """
    Remove logger from logging module.

    :param loggers_or_names: Loggers or names of loggers to remove.
    :return: None
    """

    for logger_or_name in loggers_or_names:
        if isinstance(logger_or_name, logging.Logger):
            logger_name = logger_or_name.name
        elif isinstance(logger_or_name, str):
            logger_name = logger_or_name
        else:
            raise TypeError(f"Expected 'str' or 'logging.Logger', got '{type(logger_or_name)}'.")
        if not logger_exists(logger_name):
            continue
        logging.root.manager.loggerDict.pop(logger_name)



from datetime import datetime
from typing import Optional, Any

from mongoengine import Document, StringField, DateTimeField, EmbeddedDocumentField, ReferenceField, ListField, ImageField, ImageGridFsProxy
from starlette.requests import Request

from wiederverwendbar.mongoengine.security.hashed_password import HashedPasswordDocument
from wiederverwendbar.starlette_admin.mongoengine.auth.settings import MongoengineAdminAuthSettings


class User(Document):
    class HashedPasswordDocument(HashedPasswordDocument):
        def __str__(self):
            return ""

    meta = {"collection": "user"}

    username: str = StringField(min_length=3, max_length=32, required=True, unique=True)
    password_doc: HashedPasswordDocument = EmbeddedDocumentField(HashedPasswordDocument)
    password_change_time: Optional[datetime] = DateTimeField()
    password_expiration_time: Optional[datetime] = DateTimeField()
    sessions: list[Any] = ListField(ReferenceField("Session"))
    avatar: ImageGridFsProxy = ImageField(size=(100, 100))
    company_logo: str = StringField()

    def __init__(self, *args, **values):
        super().__init__(*args, **values)

        # cleanup sessions they don't exist anymore
        sessions = []
        for session in self.sessions:
            if type(session) is not self.session_document_cls:
                continue
            sessions.append(session)
        if len(sessions) != len(self.sessions):
            self.sessions = sessions
            self.save()

    async def __admin_repr__(self, request: Request):
        return f"{self.username}"

    @property
    def password(self) -> HashedPasswordDocument:
        return self.password_doc

    @password.setter
    def password(self, value: str) -> None:
        # set password
        self.password_doc = HashedPasswordDocument.hash_password(value)

        # set password change time
        self.password_change_time = datetime.now()

        # reset password expiration time
        if self.password_expiration_time is not None:
            if self.password_change_time > self.password_expiration_time:
                self.password_expiration_time = None

    @property
    def session_document_cls(self) -> type[Any]:
        return getattr(self, "_fields")["sessions"].field.document_type_obj

    def create_session_from_request(self, request: Request) -> Any:
        # get settings
        settings = MongoengineAdminAuthSettings.from_state(state=request.state)

        # get user-agent
        user_agent = request.headers.get("User-Agent", "")

        session = self.session_document_cls(user=self,
                                            app_name=settings.admin_name,
                                            user_agent=user_agent,
                                            last_access=datetime.now())
        session.save()
        self.sessions.append(session)
        self.save()

        return session



from wiederverwendbar.mongoengine.fields import (BooleanAlsoField,
                                                 DomainField,
                                                 IPv4AddressField,
                                                 IPv4NetworkField,
                                                 PortField,
                                                 WithInstanceField)
from wiederverwendbar.mongoengine.logger import (MongoengineLogDocument,
                                                 MongoengineLogFormatter,
                                                 MongoengineLogHandler,
                                                 MongoengineLogStreamer,
                                                 mongoengine_log_stream_print)
from wiederverwendbar.mongoengine.security import (HashedPasswordDocument)
from wiederverwendbar.mongoengine.automatic_reference import (AutomaticReferenceDocument)
from wiederverwendbar.mongoengine.backup import (dump, restore)
from wiederverwendbar.mongoengine.db import (MongoengineDb)
from wiederverwendbar.mongoengine.property_document import (PropertyDocument)
from wiederverwendbar.mongoengine.settings import (MongoengineSettings)
from wiederverwendbar.mongoengine.singleton import (MongoengineDbSingleton)



from wiederverwendbar.starlette_admin import FormMaxFieldsAdmin, ActionLogAdmin, GenericEmbeddedAdmin
# from wiederverwendbar.starlette_admin.action_log.admin import ActionLogAdminMeta
# from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.admin import GenericEmbeddedAdminMeta
#
# class AdminMeta(ActionLogAdminMeta, GenericEmbeddedAdminMeta):
#     ...


class Admin(ActionLogAdmin, GenericEmbeddedAdmin, FormMaxFieldsAdmin):#, metaclass=AdminMeta):
    ...


if __name__ == "__main__":
    ...



import ctypes
import logging
import sys
import threading
import time
import traceback
from collections.abc import Callable
from contextlib import contextmanager
from datetime import datetime
from typing import Optional, Any, Union

from wiederverwendbar.default import Default
from wiederverwendbar.functions.datetime import local_now


class ThreadInterrupt(threading.ThreadError):
    """
    Exception to interrupt a thread.
    """

    ...


class ThreadLoopContinue(threading.ThreadError):
    """
    Exception to continue a loop in a thread.
    """

    ...


class ThreadStop(threading.ThreadError):
    """
    Exception to stop a thread.
    """

    ...


class ThreadKill(threading.ThreadError):
    """
    Exception to kill a thread.
    """

    ...


class ThreadWatchdogError(threading.ThreadError):
    """
    Exception to indicate an error in the watchdog of a thread.
    """

    ...


class ExtendedThread(threading.Thread):
    """
    Extended thread class with additional features.

    Features:
    - Logging
    - Interrupt handling
    - Loop handling
    - Stop handling
    - Kill handling
    - Watchdog
    - Auto start
    - Context manager for ignore
    - Context manager for loop wait
    - Thread safe properties
    """

    def __init__(self,
                 group=None,
                 target: Union[Callable[..., Any], None, Default] = Default(),
                 name: Union[str, Default] = Default(),
                 args: Union[tuple[Any, ...], Default] = Default(),
                 kwargs: Union[dict[str, Any], Default] = Default(),
                 *,
                 daemon: Union[bool, None, Default] = Default(),
                 cls_name: Union[str, Default] = Default(),
                 logger: Union[logging.Logger, Default] = Default(),
                 ignore_stop: Union[bool, Default] = Default(),
                 loop_disabled: Union[bool, Default] = Default(),
                 loop_sleep_time: Union[float, None, Default] = Default(),
                 loop_stop_on_other_exception: Union[bool, Default] = Default(),
                 continue_exceptions: Union[list[type[BaseException]], Default] = Default(),
                 stop_exceptions: Union[list[type[BaseException]], Default] = Default(),
                 kill_exceptions: Union[list[type[BaseException]], Default] = Default(),
                 watchdog_target: Union[Callable[["ExtendedThread"], bool], None, Default] = Default(),
                 auto_start: Union[bool, Default] = Default()):
        """
        Initialize the extended thread.

        :param group: Thread group.
        :param target: Thread target.
        :param name: Thread name.
        :param args: Thread arguments.
        :param kwargs: Thread keyword arguments.
        :param daemon: Thread daemon.
        :param cls_name: Class name used in the logger.
        :param logger: The logger of the thread.
        :param ignore_stop: If the stop should be ignored.
        :param loop_disabled: If the loop is disabled.
        :param loop_sleep_time: Loop sleep time.
        :param loop_stop_on_other_exception: If the loop should stop on other exceptions.
        :param continue_exceptions: Exceptions to continue the loop.
        :param stop_exceptions: Exceptions to stop the thread.
        :param kill_exceptions: Exceptions to kill the thread.
        :param watchdog_target: Watchdog target.
        :param auto_start: If the thread should start automatically.
        """

        if type(target) is Default:
            target = None

        if type(name) is Default:
            name = getattr(threading, "_newname")(f"{self.__class__.__name__}-%d")
            if target is not None:
                try:
                    target_name = target.__name__
                    name += f" ({target_name})"
                except AttributeError:
                    pass

        if type(args) is Default:
            args = ()

        if type(kwargs) is Default:
            kwargs = {}

        if type(daemon) is Default:
            daemon = None

        super().__init__(
            group=group,
            target=target,
            name=name,
            args=args,
            kwargs=kwargs,
            daemon=daemon
        )
        self.lock = threading.Lock()

        # set class name
        if type(cls_name) is Default:
            cls_name = self.__class__.__name__
        self._cls_name: str = cls_name

        # set logger
        if type(logger) is Default:
            # create a logger if none is provided
            logger = logging.getLogger(self.name)
        self._logger: logging.Logger = logger

        # set ignore stop
        if type(ignore_stop) is Default:
            ignore_stop = False
        self._ignore_stop: bool = ignore_stop

        # set loop disabled
        if type(loop_disabled) is Default:
            loop_disabled = False
        self._loop_disabled: bool = loop_disabled

        # set loop delay
        if type(loop_sleep_time) is Default:
            loop_sleep_time = None
        self._loop_sleep_time: Optional[float] = loop_sleep_time

        # set loop stop on other exception
        if type(loop_stop_on_other_exception) is Default:
            loop_stop_on_other_exception = False
        self._loop_stop_on_other_exception: bool = loop_stop_on_other_exception

        # set continue exceptions
        if type(continue_exceptions) is Default:
            continue_exceptions = []
        if ThreadLoopContinue not in continue_exceptions:
            continue_exceptions.append(ThreadLoopContinue)
        self._continue_exceptions: tuple[type[BaseException]] = tuple(continue_exceptions)

        # set stop exceptions
        if type(stop_exceptions) is Default:
            stop_exceptions = []
        if ThreadStop not in stop_exceptions:
            stop_exceptions.append(ThreadStop)
        self._stop_exceptions: tuple[type[BaseException]] = tuple(stop_exceptions)

        # set kill exceptions
        if type(kill_exceptions) is Default:
            kill_exceptions = []
        if ThreadKill not in kill_exceptions:
            kill_exceptions.append(ThreadKill)
        self._kill_exceptions: tuple[type[BaseException]] = tuple(kill_exceptions)

        # set watchdog target
        if type(watchdog_target) is Default:
            watchdog_target = None
        self._watchdog_target: Optional[Callable[[Union["ExtendedThread", Any]], bool]] = watchdog_target

        # set auto start
        if type(auto_start) is Default:
            auto_start = True
        self._auto_start: bool = auto_start

        # set internal variables
        self._started_at: Optional[datetime] = None
        self._ended_at: Optional[datetime] = None
        self._loop_started_at: Optional[datetime] = None
        self._loop_ended_at: Optional[datetime] = None
        self._loop_delay: float = 0.0
        self._wait: bool = False
        self._interrupt_exception: Optional[BaseException] = None
        self._watchdog_thread: Optional[threading.Thread] = None

        if self._auto_start:
            self.start()

    def __del__(self):
        if self.is_alive():
            self.stop()

    @property
    def logger(self) -> logging.Logger:
        """
        Get the logger of the thread.

        :rtype: logging.Logger
        :return: The logger of the thread.
        """

        with self.lock:
            return self._logger

    @logger.setter
    def logger(self, value: logging.Logger):
        """
        Set the logger of the thread.

        :param value: The new logger.
        :rtype: None
        """

        with self.lock:
            self._logger = value

    @property
    def started_at(self) -> Optional[datetime]:
        """
        Get the time when the thread was started.

        :rtype: datetime.datetime
        :return: The time when the thread was started.
        """

        with self.lock:
            return self._started_at

    @property
    def loop_started_at(self) -> Optional[datetime]:
        """
        Get the time when the loop was started.

        :rtype: datetime.datetime
        :return: The time when the loop was started.
        """

        with self.lock:
            return self._loop_started_at

    @property
    def loop_ended_at(self) -> Optional[datetime]:
        """
        Get the time when the loop was ended.

        :rtype: datetime.datetime
        :return: The time when the loop was ended.
        """

        with self.lock:
            return self._loop_ended_at

    @property
    def ended_at(self) -> Optional[datetime]:
        """
        Get the time when the thread was ended.

        :rtype: datetime.datetime
        :return: The time when the thread was ended.
        """

        with self.lock:
            return self._ended_at

    @property
    def ignore_stop(self) -> bool:
        """
        If the stop should be ignored.

        :rtype: bool
        :return: True if the stop should be ignored.
        """

        with self.lock:
            return self._ignore_stop

    @ignore_stop.setter
    def ignore_stop(self, value: bool):
        """
        Set if the stop should be ignored.

        :param value: If the stop should be ignored.
        :rtype: None
        """

        with self.lock:
            self._ignore_stop = value

    @property
    def loop_disabled(self) -> bool:
        """
        If the loop is disabled.

        :rtype: bool
        :return: True if the loop is disabled.
        """

        with self.lock:
            return self._loop_disabled

    @property
    def loop_sleep_time(self) -> Optional[float]:
        """
        Get the loop sleep time.

        :rtype: float
        :return: The loop sleep time.
        """

        with self.lock:
            return self._loop_sleep_time

    @loop_sleep_time.setter
    def loop_sleep_time(self, value: Optional[float]):
        """
        Set the loop sleep time.

        :param value: The loop sleep time.
        :rtype: None
        """

        with self.lock:
            self._loop_sleep_time = value

    @property
    def loop_delay(self) -> float:
        """
        Get the loop delay.

        :rtype: float
        :return: The loop delay.
        """

        with self.lock:
            return self._loop_delay

    @property
    def loop_stop_on_other_exception(self) -> bool:
        """
        If the loop should stop on other exceptions.

        :rtype: bool
        :return: True if the loop should stop on other exceptions.
        """

        with self.lock:
            return self._loop_stop_on_other_exception

    @loop_stop_on_other_exception.setter
    def loop_stop_on_other_exception(self, value: bool):
        """
        Set if the loop should stop on other exceptions.

        :param value: If the loop should stop on other exceptions.
        :rtype: None
        """

        with self.lock:
            self._loop_stop_on_other_exception = value

    @property
    def wait(self) -> bool:
        """
        If the thread should wait.

        :rtype: bool
        :return: True if the thread should wait.
        """

        with self.lock:
            return self._wait

    @property
    def sleep_time(self) -> Optional[float]:
        if self.loop_sleep_time is None:
            return None
        sleep_time = self.loop_sleep_time - self.loop_delay
        if sleep_time < 0:
            return 0.0
        return sleep_time

    @property
    def args(self) -> tuple[tuple[Any, ...]]:
        """
        Get the arguments of the thread.

        :rtype: tuple[tuple[Any, ...]]
        :return: The arguments of the thread.
        """

        with self.lock:
            return getattr(self, "_args", ())

    @args.setter
    def args(self, value: tuple[Any, ...]):
        """
        Set the arguments of the thread.

        :param value: The arguments of the thread.
        :rtype: None
        """

        with self.lock:
            setattr(self, "_args", value)

    @property
    def kwargs(self) -> dict[str, Any]:
        """
        Get the keyword arguments of the thread.

        :rtype: dict[str, Any]
        :return: The keyword arguments of the thread.
        """

        with self.lock:
            return getattr(self, "_kwargs", {})

    @kwargs.setter
    def kwargs(self, value: dict[str, Any]):
        """
        Set the keyword arguments of the thread.

        :param value: The keyword arguments of the thread.
        :rtype: None
        """

        with self.lock:
            setattr(self, "_kwargs", value)

    @property
    def target(self) -> Optional[Callable[..., Any]]:
        """
        Get the target of the thread.

        :rtype: Callable[..., _Any]
        :return: The target of the thread.
        """

        with self.lock:
            return getattr(self, "_target", None)

    @contextmanager
    def ignore(self) -> None:
        """
        Context manager to ignore the stop.

        :rtype: None
        :return: Nothing
        """

        ignore_stop_before = self.ignore_stop
        self.ignore_stop = True
        yield
        self.ignore_stop = ignore_stop_before

    @contextmanager
    def loop_wait(self, block: bool = True, timeout: Optional[float] = None) -> None:
        """
        Context manager to wait for the next loop.

        :param block: Block before entering the context manager.
        :param timeout: Timeout for the block.
        :rtype: None
        :return: Nothing
        """

        loop_wait_before = self._wait
        with self.lock:
            self._wait = True
        if block:  # wait for next loop
            loop_started_at = self.loop_started_at
            if loop_started_at is not None:
                block_start_counter = time.perf_counter()
                while loop_started_at != self.loop_started_at:
                    if timeout is not None:
                        if time.perf_counter() - block_start_counter > timeout:
                            raise TimeoutError("Timeout while waiting for loop.")
                    time.sleep(0.001)
        yield
        with self.lock:
            self._wait = loop_wait_before

    def start_watchdog(self) -> None:
        """
        Start the watchdog. If the watchdog is already running, nothing happens.

        :rtype: None
        :return: Nothing
        """

        if self._watchdog_target is None:
            return  # watchdog is disabled
        if self._watchdog_thread is not None:
            if self._watchdog_thread.is_alive():
                return  # watchdog is already running
        self._watchdog_thread = threading.Thread(name=f"{self.name}.watchdog", target=self._watchdog_loop, daemon=True)
        self._watchdog_thread.start()

    def _watchdog_loop(self) -> None:
        """
        Watchdog of the thread.

        :rtype: None
        :return: Nothing
        """

        self.logger.debug(f"{self._cls_name} watchdog started.")

        while True:
            watchdog_loop_start_counter = time.perf_counter()
            try:
                watchdog_target_result = bool(self._watchdog_target(self))
                if not watchdog_target_result:
                    self.logger.info(f"{self._cls_name} watchdog received stop signal.")
                    break
            except BaseException as e:
                handle_exception(msg=f"{self._cls_name} watchdog raised an exception", e=e, logger=self.logger, chain=False)
                self.raise_exception(ThreadWatchdogError)
            watchdog_loop_delay = time.perf_counter() - watchdog_loop_start_counter
            if self.loop_sleep_time:
                sleep_start_counter = time.perf_counter()
                while time.perf_counter() - sleep_start_counter < self.loop_sleep_time - watchdog_loop_delay:
                    time.sleep(0.001)


        self.logger.debug(f"{self._cls_name} watchdog ended.")

    def raise_exception(self, exception: Union[type[BaseException], BaseException]) -> None:
        """
        Raises the given exception in the context of this thread.

        :param exception: The exception to raise.
        :rtype: None
        :return: Nothing
        """

        # check if the exception is an Exception type
        if isinstance(exception, BaseException):
            self._interrupt_exception = exception
            exception = ThreadInterrupt
        elif issubclass(exception, BaseException):
            self._interrupt_exception = None
        else:
            raise TypeError("Only types or object derived from BaseException can be raised")

        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.ident), ctypes.py_object(exception))
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # if it returns a number greater than one, you're in trouble, and you should call it again with exc=None to revert the effect
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self.ident), None)
            raise SystemError("PyThreadState_SetAsyncExc failed")

    def stop(self) -> None:
        """
        Send a stop signal to the thread.

        :rtype: None
        :return: Nothing
        """

        self.raise_exception(ThreadStop)

    def kill(self) -> None:
        """
        Send a kill signal to the thread.

        :rtype: None
        :return: Nothing
        """

        self.raise_exception(ThreadKill)

    def run(self) -> None:
        """
        THe inner run method of the thread. Don't override this method. Use loop instead.

        :rtype: None
        :return: Nothing
        """

        self.on_start()

        with self.lock:
            self._started_at = local_now()

        self.logger.info(f"{self._cls_name} started.")

        while True:
            # start watchdog
            self.start_watchdog()

            try:
                try:
                    # get loop start time
                    with self.lock:
                        self._loop_started_at = local_now()
                    loop_start_counter = time.perf_counter()

                    self.logger.debug(f"{self._cls_name} is running loop.")

                    # wait
                    if self.wait:
                        self.logger.debug(f"{self._cls_name} loop is waiting.")
                        while self.wait:
                            time.sleep(0.001)
                        self.logger.debug(f"{self._cls_name} loop is continuing.")

                    # execute loop start
                    self.on_loop_start()

                    # execute loop
                    self.loop()

                    # execute loop end
                    self.on_loop_end()

                    with self.lock:
                        # set loop end time
                        self._loop_ended_at = local_now()

                        # set loop delay
                        self._loop_delay = time.perf_counter() - loop_start_counter

                    # sleep if necessary
                    if not self.loop_disabled:
                        if self.sleep_time:
                            self.logger.debug(f"{self._cls_name} loop is sleeping for {self.sleep_time} seconds.")
                            sleep_start_counter = time.perf_counter()
                            while time.perf_counter() - sleep_start_counter < self.sleep_time:
                                time.sleep(0.001)
                    else:
                        break
                except ThreadInterrupt:
                    if self._interrupt_exception is None:
                        raise RuntimeError("ThreadInterrupt was raised but no exception was set.")
                    else:
                        raise self._interrupt_exception
            except self._continue_exceptions as e:
                self.logger.debug(f"{self._cls_name} received {e.__class__.__name__}. Continue loop.")
                continue
            except self._stop_exceptions as e:
                if self.ignore_stop:
                    self.logger.debug(f"{self._cls_name} received {e.__class__.__name__} but ignore_stop is True. Continue loop.")
                    continue
                else:
                    self.logger.debug(f"{self._cls_name} received {e.__class__.__name__}. Stop loop.")
                    self.on_stop()
                    break
            except self._kill_exceptions as e:
                self.logger.debug(f"{self._cls_name} received {e.__class__.__name__}. Kill loop.")
                break
            except BaseException as e:
                if self._interrupt_exception is None:
                    handle_exception(msg=f"{self._cls_name} loop raised an exception", e=e, logger=self.logger, chain=True)
                else:
                    self._interrupt_exception = None
                    handle_exception(msg=f"{self._cls_name} loop raised an exception", e=e, logger=self.logger, chain=False)

                if self._loop_stop_on_other_exception:
                    break

        # execute end
        self.on_end()

        with self.lock:
            self._ended_at = local_now()

        self.logger.info(f"{self._cls_name} ended.")

    def on_start(self) -> None:
        """
        Method to execute on start. This method is called before the loop starts by the run method.
        You can override this method.

        :rtype: None
        :return: Nothing
        """

        ...

    def on_loop_start(self) -> None:
        """
        Method to execute on loop start. This method is called every time the loop starts by the run method.
        You can override this method.

        :rtype: None
        :return: Nothing
        """

        ...

    def on_loop_end(self) -> None:
        """
        Method to execute on loop end. This method is called every time the loop ends by the run method.
        You can override this method.

        :rtype: None
        :return: Nothing
        """

        ...

    def on_stop(self) -> None:
        """
        Method to execute on stop. This method is called when the thread is stopped by the run method.
        You can override this method.

        :rtype: None
        :return: Nothing
        """

        ...

    def on_end(self) -> None:
        """
        Method to execute on end. This method is called when the thread stopped or killed by the run method.
        You can override this method.

        :rtype: None
        :return:
        """

        ...

    def loop(self) -> None:
        """
        The loop method. You can override this method.

        :rtype: None
        :return: Nothing
        """

        if self.target is None:
            return
        self.target(*self.args, **self.kwargs)


def handle_exception(msg: str, e: BaseException, logger: logging.Logger, chain: bool = True) -> str:
    """
    Handle an exception.

    :param msg: The name of the exception.
    :param e: The exception.
    :param logger: The logger.
    :param chain: If the exception chain should be printed.
    :rtype: str
    :return: The exception message.
    """

    tb_str = "".join(traceback.format_exception(type(e), value=e, tb=e.__traceback__, chain=chain)).strip()
    msg += f":\n{tb_str}"
    logger.error(msg)
    print(msg, file=sys.stderr)
    return msg



from wiederverwendbar.starlette_admin.action_log import (WebsocketHandler,
                                                         ActionLoggerCommand,
                                                         ActionLoggerResponse,
                                                         ActionSubLogger,
                                                         ActionSubLoggerContext,
                                                         ActionLogger,
                                                         ActionLogAdmin,
                                                         ActionThread,
                                                         ExitCommand,
                                                         FormCommand,
                                                         FinalizeCommand,
                                                         IncreaseStepsCommand,
                                                         NextStepCommand,
                                                         StepCommand,
                                                         ConfirmCommand,
                                                         DownloadCommand,
                                                         YesNoCommand)

from wiederverwendbar.starlette_admin.drop_down_icon_view import (DropDownIconViewAdmin,
                                                                  DropDownIconView)

from wiederverwendbar.starlette_admin.mongoengine import (AuthView,
                                                          Session,
                                                          SessionView,
                                                          User,
                                                          UserView,
                                                          MongoengineAuthAdmin,
                                                          MongoengineAdminAuthProvider,
                                                          MongoengineAdminAuthSettings,
                                                          BooleanAlsoAdmin,
                                                          BooleanAlsoConverter,
                                                          BooleanAlsoField,
                                                          GenericEmbeddedAdmin,
                                                          GenericEmbeddedConverter,
                                                          GenericEmbeddedDocumentField,
                                                          ListField,
                                                          GenericEmbeddedDocumentView,
                                                          IPv4Converter,
                                                          FixedModelView,
                                                          MongoengineModelView,
                                                          MongoengineConverter)

from wiederverwendbar.starlette_admin.form_max_fields import (FormMaxFieldsAdmin,
                                                              FormMaxFieldsAdminSettings)

from wiederverwendbar.starlette_admin.multi_path import (MultiPathAdmin)

from wiederverwendbar.starlette_admin.settings import (SettingsAdmin,
                                                       AdminSettings)

from wiederverwendbar.starlette_admin.admin import (BaseAdmin)



from ipaddress import IPv4Address
from pathlib import Path
from typing import Optional, Union

from wiederverwendbar.pydantic.printable_settings import PrintableSettings, Field


class SqlalchemySettings(PrintableSettings):
    db_file: Optional[Path] = Field(default=None,
                                    title="Database File",
                                    description="File to connect to database")
    db_host: Union[IPv4Address, str, None] = Field(default=None,
                                                   title="Database Host",
                                                   description="Host to connect to database")

    db_port: Optional[int] = Field(default=None,
                                   title="Database Port",
                                   ge=0,
                                   le=65535,
                                   description="Port to connect to database")
    db_protocol: str = Field(default="sqlite",
                             title="Database Protocol",
                             description="Protocol to connect to database")
    db_name: Optional[str] = Field(default=None,
                                   title="Database Name",
                                   description="Name of the database")
    db_username: Optional[str] = Field(default=None,
                                       title="Database User",
                                       description="User to connect to database")
    db_password: Optional[str] = Field(None,
                                       title="Database Password",
                                       description="Password to connect to database",
                                       secret=True)
    db_echo: bool = Field(default=False,
                          title="Database echo.",
                          description="Echo SQL queries to console")



from typing import Union, Optional

from mongoengine import GridFSProxy, ImageGridFsProxy
from starlette.requests import Request


def get_grid_fs_url(file: Union[GridFSProxy, ImageGridFsProxy], request: Request, use_thumbnail: bool = True) -> Optional[str]:
    grid_id = file.grid_id
    if use_thumbnail and getattr(file, "thumbnail_id", None):
        grid_id = file.thumbnail_id
    if grid_id is None:
        return None
    url = request.url_for(
        request.app.state.ROUTE_NAME + ":api:file",
        db=file.db_alias,
        col=file.collection_name,
        pk=grid_id,
    )
    url_str = str(url)
    return url_str



from enum import Enum


class LogLevels(str, Enum):
    """
    Log levels
    """

    CRITICAL = "CRITICAL"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"



from mongoengine import StringField


class DomainField(StringField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, regex=r"^(?:$|(?=.{1,253}$)(?:(?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,63})$", **kwargs)



from enum import Enum

from pydantic import Field

from wiederverwendbar.starlette_admin.settings.settings import AdminSettings


class ActionLogAdminSettings(AdminSettings):
    class LogLevels(str, Enum):
        """
        Log levels
        """

        CRITICAL = "CRITICAL"
        FATAL = "FATAL"
        ERROR = "ERROR"
        WARNING = "WARNING"
        INFO = "INFO"
        DEBUG = "DEBUG"

    action_log_level: LogLevels = Field(default=LogLevels.INFO,
                                        title="Action Log Level",
                                        description="The action log level")
    action_log_ignore_loggers_equal: list[str] = Field(default_factory=list,
                                                       title="Action Log Ignore Loggers Equal",
                                                       description="List of loggers to ignore equal")
    action_log_ignore_loggers_like: list[str] = Field(default_factory=list,
                                                      title="Action Log Ignore Loggers Like",
                                                      description="List of loggers to ignore like")
    action_log_formatter: str = Field(default="%(asctime)s - %(levelname)s - %(message)s",
                                      title="Action Log Formatter",
                                      description="The action log formatter")
    action_log_wait_for_websocket: bool = Field(default=True,
                                                title="Action Log Wait For WebSocket",
                                                description="Wait for the websocket to be ready")
    action_log_wait_for_websocket_timeout: int = Field(default=5,
                                                       title="Action Log Wait For WebSocket Timeout",
                                                       description="The action log wait for websocket timeout")
    action_log_show_errors: bool = Field(default=True,
                                         title="Action Log Show Errors",
                                         description="Show errors in the action log")
    action_log_halt_on_error: bool = Field(default=True,
                                           title="Action Log Halt On Error",
                                           description="Halt on error in the action log")
    action_log_use_context_logger_level: bool = Field(default=True,
                                                      title="Action Log Use Context Logger Level",
                                                      description="Use the context logger level in the action log")
    action_log_use_context_logger_level_on_not_set: bool = Field(default=True,
                                                                 title="Action Log Use Context Logger Level On Not Set",
                                                                 description="Use the context logger level on not set in the action log")
    action_log_handle_origin_logger: bool = Field(default=True,
                                                  title="Action Log Handle Origin Logger",
                                                  description="Handle the origin logger in the action log")



from enum import Enum

from bson import ObjectId
from mongoengine import signals, Document, ListField, ReferenceField
from mongoengine.base import TopLevelDocumentMetaclass


class AutomaticReferenceTypes(Enum):
    ONE_TO_ONE = "one_to_one"
    ONE_TO_MANY = "one_to_many"
    MANY_TO_ONE = "many_to_one"
    MANY_TO_MANY = "many_to_many"


class AutomaticReferenceDocumentMeta(TopLevelDocumentMetaclass):
    def __new__(mcs, name, bases, attrs):
        # add attribute '_automatic_reference_fields' to class for caching
        attrs["_automatic_reference_fields"] = None

        # create new class
        new_class = super().__new__(mcs, name, bases, attrs)

        # connect signals
        signals.post_save.connect(mcs.post_save, sender=new_class)
        signals.pre_delete.connect(mcs.pre_delete, sender=new_class)

        return new_class

    @classmethod
    def get_automatic_reference_fields(cls, document) -> list[tuple[str, "AutomaticReferenceDocumentMeta", AutomaticReferenceTypes, str]]:
        # return cached reference fields
        automatic_reference_fields = getattr(document, "_automatic_reference_fields")
        if automatic_reference_fields is not None:
            return automatic_reference_fields

        # get all reference fields
        all_fields = getattr(document, "_fields")
        automatic_reference_fields = []
        for field_name in getattr(document, "_fields_ordered"):
            field = all_fields[field_name]
            automatic_reference = getattr(field, "automatic_reference", None)
            if automatic_reference is None:
                continue
            if not isinstance(field, ReferenceField) and not isinstance(field, ListField):
                continue

            # get reference field
            reference_field = None
            if isinstance(field, ReferenceField):
                reference_field = field
            elif isinstance(field, ListField):
                if not hasattr(field, "field"):
                    raise ValueError(f"ListField {field.name} has no field attribute.")
                if not isinstance(field.field, ReferenceField):
                    raise ValueError(f"ListField {field.name} has no ReferenceField as field attribute.")
                reference_field = field.field
            if reference_field is None:
                raise ValueError(f"Document type of field {field.name} can not be resolved.")

            # get document type
            document_type = getattr(reference_field, "document_type")
            if not isinstance(document_type, AutomaticReferenceDocumentMeta):
                raise ValueError(f"config field of {cls.__name__} can not be resolved to a document type.")

            # get automatic reference type
            reference_type = None
            field_from_reference = getattr(document_type, automatic_reference)
            field_from_reference_automatic_reference = getattr(field_from_reference, "automatic_reference")
            if field_from_reference_automatic_reference != field_name:
                raise ValueError(f"Automatic reference field {field_name} not found in document {document_type.__name__}.{automatic_reference}.")
            if isinstance(field, ReferenceField):
                if isinstance(field_from_reference, ReferenceField):  # ONE_TO_ONE
                    reference_type = AutomaticReferenceTypes.ONE_TO_ONE
                elif isinstance(field_from_reference, ListField):  # ONE_TO_MANY
                    reference_type = AutomaticReferenceTypes.ONE_TO_MANY
            elif isinstance(field, ListField):
                if isinstance(field_from_reference, ReferenceField):  # MANY_TO_ONE
                    reference_type = AutomaticReferenceTypes.MANY_TO_ONE
                elif isinstance(field_from_reference, ListField):  # MANY_TO_MANY
                    reference_type = AutomaticReferenceTypes.MANY_TO_MANY
            if reference_type is None:
                raise ValueError(f"Field type for field {field.name} not found.")

            automatic_reference_fields.append((field_name, document_type, reference_type, automatic_reference))

        # cache reference fields
        setattr(document, "_automatic_reference_fields", automatic_reference_fields)

        return automatic_reference_fields

    @classmethod
    def resolve_object_id(cls, document, field_name, field, reference_document):
        if type(field) is ObjectId:
            field = reference_document.objects(id=field).first()
            if field is None:
                raise ValueError(f"{reference_document.__name__} with id {getattr(document, field_name)} not found.")
            setattr(document, field_name, field)
        return field

    @classmethod
    def merge_and_save(cls, save_documents):
        # merge all documents
        merged_documents = []
        for reference_document, documents in save_documents.items():
            merged_documents_by_id = {}
            for document in documents:
                if document.id not in merged_documents_by_id:
                    merged_documents_by_id[document.id] = document
                else:
                    for changed_field in getattr(document, "_changed_fields"):
                        setattr(merged_documents_by_id[document.id], changed_field, getattr(document, changed_field))
            merged_documents.extend(merged_documents_by_id.values())

        # save merged documents
        for document in merged_documents:
            document.save()

    @classmethod
    def post_save(cls, sender, document, **kwargs):
        save_documents = {}
        # save all reference fields
        for field_name, reference_document, reference_type, reference_field_name in cls.get_automatic_reference_fields(document):
            if reference_document not in save_documents:
                save_documents[reference_document] = []
            if reference_type == AutomaticReferenceTypes.ONE_TO_ONE:
                save_documents[reference_document].extend(cls.save_one_to_one(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.ONE_TO_MANY:
                save_documents[reference_document].extend(cls.save_one_to_many(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.MANY_TO_ONE:
                save_documents[reference_document].extend(cls.save_many_to_one(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.MANY_TO_MANY:
                save_documents[reference_document].extend(cls.save_many_to_many(document, field_name, reference_document, reference_field_name))
            else:
                raise ValueError(f"Reference type {reference_type} not not supported.")

        # merge and save all documents
        cls.merge_and_save(save_documents)

    @classmethod
    def save_one_to_one(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # add document to reference field
        field = getattr(document, field_name)
        if field is not None:
            reference = cls.resolve_object_id(document, field_name, field, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document != reference_field:
                setattr(reference, reference_field_name, document)
                save_references.append(reference)
        # remove document from old reference field
        for reference in reference_document.objects(**{reference_field_name: document}):
            if field == reference:
                continue
            setattr(reference, reference_field_name, None)
            save_references.append(reference)
        return save_references

    @classmethod
    def save_one_to_many(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # add document to reference field
        field = getattr(document, field_name)
        if field is not None:
            reference = cls.resolve_object_id(document, field_name, field, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document not in reference_field:
                reference_field.append(document)
                setattr(reference, reference_field_name, reference_field)
                save_references.append(reference)
        # remove document from old reference field
        for reference in reference_document.objects(**{reference_field_name: document}):
            if reference == field:
                continue
            reference_field = getattr(reference, reference_field_name)
            reference_field.remove(document)
            setattr(reference, reference_field_name, reference_field)
            save_references.append(reference)
        return save_references

    @classmethod
    def save_many_to_one(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # add document to reference field
        field = getattr(document, field_name)
        for reference in field:
            reference = cls.resolve_object_id(document, field_name, reference, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document != reference_field:
                setattr(reference, reference_field_name, document)
                save_references.append(reference)
        # remove document from old reference field
        for reference in reference_document.objects(**{reference_field_name: document}):
            if reference in field:
                continue
            setattr(reference, reference_field_name, None)
            save_references.append(reference)
        return save_references

    @classmethod
    def save_many_to_many(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # add document to reference field
        field = getattr(document, field_name)
        for reference in field:
            reference = cls.resolve_object_id(document, field_name, reference, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document not in reference_field:
                reference_field.append(document)
                setattr(reference, reference_field_name, reference_field)
                save_references.append(reference)
        # remove document from old reference field
        for reference in reference_document.objects(**{reference_field_name: document}):
            if reference in field:
                continue
            reference_field = getattr(reference, reference_field_name)
            reference_field.remove(document)
            setattr(reference, reference_field_name, reference_field)
            save_references.append(reference)
        return save_references

    @classmethod
    def pre_delete(cls, sender, document, **kwargs):
        save_documents = {}
        # delete all reference fields
        for field_name, reference_document, reference_type, reference_field_name in cls.get_automatic_reference_fields(document):
            if reference_document not in save_documents:
                save_documents[reference_document] = []
            if reference_type == AutomaticReferenceTypes.ONE_TO_ONE:
                save_documents[reference_document].extend(cls.delete_one_to_one(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.ONE_TO_MANY:
                save_documents[reference_document].extend(cls.delete_one_to_many(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.MANY_TO_ONE:
                save_documents[reference_document].extend(cls.delete_many_to_one(document, field_name, reference_document, reference_field_name))
            elif reference_type == AutomaticReferenceTypes.MANY_TO_MANY:
                save_documents[reference_document].extend(cls.delete_many_to_many(document, field_name, reference_document, reference_field_name))
            else:
                raise ValueError(f"Reference type {reference_type} not not supported.")

        # merge and save all documents
        cls.merge_and_save(save_documents)

    @classmethod
    def delete_one_to_one(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # remove document from reference field
        field = getattr(document, field_name)
        if field is not None:
            reference = cls.resolve_object_id(document, field_name, field, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document == reference_field:
                setattr(reference, reference_field_name, None)
                save_references.append(reference)
        return save_references

    @classmethod
    def delete_one_to_many(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # remove document from reference field
        field = getattr(document, field_name)
        if field is not None:
            reference = cls.resolve_object_id(document, field_name, field, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document in reference_field:
                reference_field.remove(document)
                setattr(reference, reference_field_name, reference_field)
                save_references.append(reference)
        return save_references

    @classmethod
    def delete_many_to_one(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # remove document from reference field
        field = getattr(document, field_name)
        for reference in field:
            reference = cls.resolve_object_id(document, field_name, reference, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document == reference_field:
                setattr(reference, reference_field_name, None)
                save_references.append(reference)
        return save_references

    @classmethod
    def delete_many_to_many(cls, document, field_name, reference_document, reference_field_name) -> list["AutomaticReferenceDocumentMeta"]:
        save_references = []
        # remove document from reference field
        field = getattr(document, field_name)
        for reference in field:
            reference = cls.resolve_object_id(document, field_name, reference, reference_document)
            reference_field = getattr(reference, reference_field_name)
            if document in reference_field:
                reference_field.remove(document)
                setattr(reference, reference_field_name, reference_field)
                save_references.append(reference)
        return save_references


class AutomaticReferenceDocument(Document, metaclass=AutomaticReferenceDocumentMeta):
    meta = {"abstract": True}



from typing import Optional, List, Union, Type

from starlette_admin.views import BaseView, DropDown


class DropDownIconView(DropDown):
    def __init__(
            self,
            label: str,
            views: List[Union[Type[BaseView], BaseView]],
            icon: Optional[str] = None,
            always_open: bool = True,
            show_icon_on_sub_views: bool = True
    ):
        super().__init__(label=label, views=views, icon=icon, always_open=always_open)

        self.show_icon_on_sub_views = show_icon_on_sub_views



import atexit as _atexit
import logging as _logging
import threading as _threading
from itertools import count as _count
from typing import Optional as _Optional

from wiederverwendbar.functions.datetime import local_now as _local_now
from wiederverwendbar.mongoengine.logger.documets import MongoengineLogDocument as _MongoengineLogDocument
from wiederverwendbar.mongoengine.logger.formatters import MongoengineLogFormatter as _MongoengineLogFormatter

_counter = _count().__next__
_counter() # skip 0

class MongoengineLogHandler(_logging.Handler):
    def __init__(self,
                 level=_logging.NOTSET,
                 document: _Optional[type[_MongoengineLogDocument]] = None,
                 document_kwargs: _Optional[dict] = None,
                 buffer_size: _Optional[int] = None,
                 buffer_periodical_flush_timing: _Optional[float] = None,
                 buffer_early_flush_level: _Optional[int] = None):
        super().__init__(level=level)
        self.formatter: _MongoengineLogFormatter = _MongoengineLogFormatter()
        if document is None:
            document = _MongoengineLogDocument
        self._document: type[_MongoengineLogDocument] = document
        if document_kwargs is None:
            document_kwargs = {}
        self._document_kwargs: dict = document_kwargs
        self._buffer: list[_logging.LogRecord] = []
        if buffer_size is None:
            buffer_size = 100
        self._buffer_size: int = buffer_size
        if buffer_periodical_flush_timing is None:
            buffer_periodical_flush_timing = 5.0
        self._buffer_periodical_flush_timing: float = buffer_periodical_flush_timing
        if buffer_early_flush_level is None:
            buffer_early_flush_level = _logging.CRITICAL
        self._buffer_early_flush_level: int = buffer_early_flush_level
        self._buffer_timer_thread: _Optional[_threading.Thread] = None
        self._buffer_lock: _threading.Lock = _threading.Lock()
        self._stopper: _Optional[callable] = None

        # setup periodical flush
        if self._buffer_periodical_flush_timing:
            # clean exit event
            _atexit.register(self.close)

            # call at interval function
            def call_repeatedly(interval, func, *args):
                stopped = _threading.Event()

                # actual thread function
                def loop():
                    while not stopped.wait(interval):  # the first call is in `interval` secs
                        func(*args)

                timer_thread = _threading.Thread(name=f"{self.__class__.__name__}-{_counter()}", target=loop, daemon=True)
                timer_thread.start()
                return stopped.set, timer_thread

            # launch thread
            self._stopper, self._buffer_timer_thread = call_repeatedly(self._buffer_periodical_flush_timing, self.flush)

    def emit(self, record: _logging.LogRecord) -> None:
        with self._buffer_lock:
            self._buffer.append(record)

        if len(self._buffer) >= self._buffer_size or record.levelno >= self._buffer_early_flush_level:
            self.flush()

    def flush(self):
        if len(self._buffer) == 0:
            return

        with self._buffer_lock:
            formated_records = []
            for record in self._buffer:
                try:
                    formated_record = self.formatter.format(record)
                    formated_records.append(formated_record)
                except Exception:
                    self.handleError(record)
                if len(formated_records) == 0:
                    continue
            if len(formated_records) > 0:
                # create document
                document = self._document(timestamp=_local_now(),
                                          entries=formated_records,
                                          **self._document_kwargs)
                document.save()
            self.empty_buffer()

    def empty_buffer(self) -> None:
        """
        Empty the buffer list.

        :return: None
        """
        del self._buffer
        self._buffer = []

    def close(self) -> None:
        """
        Clean quit logging. Flush buffer. Stop the periodical thread if needed.

        :return: None
        """

        if self._stopper:
            self._stopper()
        self.flush()
        super().close()



from dataclasses import dataclass
from typing import List, Optional

from starlette.requests import Request
from starlette_admin import RequestAction
from starlette_admin.fields import BooleanField

@dataclass
class BooleanAlsoField(BooleanField):
    also: str = ""

    def additional_js_links(self, request: Request, action: RequestAction) -> List[str]:
        additional_js_links = super().additional_js_links(request, action)
        if action.is_form():
            additional_js_links.append(
                str(
                    request.url_for(
                        f"{request.app.state.ROUTE_NAME}:statics",
                        path="js/boolean_also_field.js",
                    )
                )
            )

        return additional_js_links



import logging
import time

from wiederverwendbar.logger import LoggerSingleton, LoggerSettings
from wiederverwendbar.mongoengine import MongoengineDbSingleton

from tests.task_manager import Manager, Worker, ScheduledTask

# init logging
LoggerSingleton(name="test", settings=LoggerSettings(log_level="DEBUG"), ignored_loggers_like=["pymongo"], init=True)

# init database
MongoengineDbSingleton(init=True)

manager = Manager()

logger = logging.getLogger(__name__)


@manager.register_task()
def task1(start: int, end: int = 10):
    # worker = Worker()
    # task = ScheduledTask()
    logger.info(f"Count from {start} to {end}")
    for i in range(start, end + 1):
        logger.info(i)
        time.sleep(1)
    logger.info("Done.")



from mongoengine import EmbeddedDocument, StringField, IntField
from wiederverwendbar.functions.security.hashed_password import HashedPassword


class HashedPasswordDocument(EmbeddedDocument, HashedPassword):
    meta = {"allow_inheritance": True}

    encoding: str = StringField(required=True)
    hash_function: str = StringField(required=True)
    interactions: int = IntField(required=True)
    key_length: int = IntField(required=True)
    salt: str = StringField(required=True)
    hashed_password: str = StringField(required=True)



import warnings
from enum import Enum
from pathlib import Path
from typing import Optional, Any

from pydantic import BaseModel, Field, DirectoryPath
from starlette.datastructures import State
from starlette.requests import Request

from starlette_admin.i18n import SUPPORTED_LOCALES


class AdminSettings(BaseModel):
    admin_title: str = Field(default="Admin", title="Admin Title", description="The title of the admin panel.")
    admin_name: str = Field(default="admin", pattern=r"^[a-zA-Z0-9_-]+$", title="Admin Name", description="The name of the admin panel.")
    admin_base_url: str = Field(default="/admin", title="Admin Base URL", description="The base URL of the admin panel.")
    admin_route_name: str = Field(default="admin", title="Admin Route Name", description="The route name of the admin panel.")
    admin_logo_url: Optional[str] = Field(default="logo.png", title="Admin Logo URL", description="The URL of the admin panel logo.")
    admin_login_logo_url: Optional[str] = Field(default="logo.png", title="Admin Login Logo URL", description="The URL of the admin panel login logo.")
    admin_templates_dir: DirectoryPath = Field(default=..., title="Admin Templates Directory", description="The directory of the admin panel templates.")
    admin_static_dir: Optional[DirectoryPath] = Field(default=None, title="Admin static Directory", description="The directory of the admin panel static.")
    admin_debug: bool = Field(default=False, title="Admin Debug", description="Whether the admin panel is in debug mode.")

    class Language(str, Enum):
        DE = "de" if "de" in SUPPORTED_LOCALES else ValueError("German is not a supported locale")
        EN = "en" if "en" in SUPPORTED_LOCALES else ValueError("English is not a supported locale")
        FR = "fr" if "fr" in SUPPORTED_LOCALES else ValueError("French is not a supported locale")
        RU = "ru" if "ru" in SUPPORTED_LOCALES else ValueError("Russian is not a supported locale")
        TR = "tr" if "tr" in SUPPORTED_LOCALES else ValueError("Turkish is not a supported locale")

    admin_language: Language = Field(default=Language.DE, title="Admin Language", description="The language of the admin panel.")
    admin_language_cookie_name: str = Field(default="language", title="Admin Language Cookie Name", description="The name of the admin panel language cookie.")
    admin_language_header_name: str = Field(default="Accept-Language", title="Admin Language Header Name", description="The name of the admin panel language header.")
    admin_language_available: Optional[list[Language]] = Field(default=None, title="Admin Language Available",
                                                               description="The available languages of the admin panel.")
    admin_favicon_url: Optional[str] = Field(default="favicon.ico", title="Admin Favicon URL", description="The URL of the admin panel favicon.")
    admin_session_secret_key: str = Field("change_me", title="Admin Session Secret Key", description="The secret key of the admin panel session.")
    admin_session_cookie: str = Field(default="session", title="Admin Session Cookie", description="The name of the admin panel session cookie.")
    admin_session_max_age: int = Field(default=14 * 24 * 60 * 60, title="Admin Session Max Age",
                                       description="The maximum age of the admin panel session. "
                                                   "If the session is not used for this time, it will be deleted.")
    admin_session_absolute_max_age: Optional[int] = Field(default=14 * 24 * 60 * 60, title="Admin Session Absolute Max Age",
                                                          description="The absolute maximum age of the admin panel session. "
                                                                      "The session will be deleted after this time. If not set, the session will not expire. "
                                                                      "If timeout_max_age is set and this is smaller than timeout_max_age, this will be set to timeout_max_age.")
    admin_session_only_one: bool = Field(default=False, title="Admin Session Only One", description="Whether only one session is allowed for the admin panel.")
    admin_session_path: str = Field(default="/", title="Admin Session Path", description="The path of the admin panel session.")

    class SameSite(str, Enum):
        LAX = "lax"
        STRICT = "strict"
        NONE = "none"

    admin_session_same_site: SameSite = Field(default=SameSite.LAX, title="Admin Session Same Site", description="The same site of the admin panel session.")
    admin_session_https_only: bool = Field(default=False, title="Admin Session HTTPS Only", description="Whether the admin panel session is HTTPS only.")
    admin_session_domain: Optional[str] = Field(None, title="Admin Session Domain", description="The domain of the admin panel session.")

    def __init__(self, /, **data: Any):
        data["admin_templates_dir"] = data.get("admin_templates_dir", Path("templates"))
        super().__init__(**data)

        # check if admin_static_dir is set
        if self.admin_static_dir is None:
            warnings.warn("Admin static directory is not set. Please set it to the directory of the admin panel static.", UserWarning)
        else:
            if not self.admin_static_dir.is_dir():
                raise FileNotFoundError(f"Admin static directory {self.admin_static_dir} is not a directory.")

            admin_logo_file = self.admin_static_dir / self.admin_logo_url
            if admin_logo_file.is_file():
                self.admin_logo_url = f"{self.admin_base_url}/statics/{self.admin_logo_url}"
            else:
                self.admin_logo_url = None
            admin_login_logo_file = self.admin_static_dir / self.admin_login_logo_url
            if admin_login_logo_file.is_file():
                self.admin_login_logo_url = f"{self.admin_base_url}/statics/{self.admin_login_logo_url}"
            else:
                self.admin_login_logo_url = None
            admin_favicon_file = self.admin_static_dir / self.admin_favicon_url
            if admin_favicon_file.is_file():
                self.admin_favicon_url = f"{self.admin_base_url}/statics/{self.admin_favicon_url}"
            else:
                self.admin_favicon_url = None

        # check if and admin_session_secret_key is not the default value
        if self.admin_session_secret_key == self.model_fields.get("admin_session_secret_key").default:
            warnings.warn("Admin session secret key is not set. Please set it to a secure value.", UserWarning)

        # check if admin_session_max_age is greater than admin_session_timeout_max_age
        if self.admin_session_absolute_max_age is not None and self.admin_session_absolute_max_age < self.admin_session_max_age:
            warnings.warn("Admin session absolute max age is smaller than admin session max age. "
                          "Setting admin session absolute max age to admin session max age.", UserWarning)
            self.admin_session_absolute_max_age = self.admin_session_max_age

    @classmethod
    def from_state(cls, state: State):
        settings = state.admin.settings
        if not isinstance(settings, cls):
            raise ValueError(f"Settings is not instance of {cls}")
        return settings


import asyncio
import logging
import threading
import time
from typing import Optional

from wiederverwendbar.starlette_admin.action_log.logger import ActionLogger


class ActionThread(threading.Thread):
    def __init__(self,
                 action_logger: ActionLogger,
                 name: str,
                 payload: Optional[callable] = None,
                 payload_args: Optional[list] = None,
                 payload_kwargs: Optional[dict] = None,
                 title: Optional[str] = None,
                 log_level: int = logging.NOTSET,
                 parent: Optional[logging.Logger] = None,
                 formatter: Optional[logging.Formatter] = None,
                 steps: Optional[int] = None,
                 on_success_msg: Optional[str] = None,
                 on_error_msg: Optional[str] = "Something went wrong.",
                 end_steps: Optional[int] = None,
                 show_errors: Optional[bool] = None,
                 halt_on_error: Optional[bool] = None,
                 use_context_logger_level: bool = True,
                 use_context_logger_level_on_not_set: Optional[bool] = None,
                 ignore_loggers_equal: Optional[list[str]] = None,
                 ignore_loggers_like: Optional[list[str]] = None,
                 handle_origin_logger: bool = True,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._payload = payload
        self._payload_args = payload_args if payload_args is not None else []
        self._payload_kwargs = payload_kwargs if payload_kwargs is not None else {}

        self._action_logger = action_logger
        self._sub_logger = None

        self._name = name
        self._title = title
        self._log_level = log_level
        self._parent = parent
        self._formatter = formatter
        self._steps = steps
        self._on_success_msg = on_success_msg
        self._on_error_msg = on_error_msg
        self._end_steps = end_steps
        self._show_errors = show_errors
        self._halt_on_error = halt_on_error
        self._use_context_logger_level = use_context_logger_level
        self._use_context_logger_level_on_not_set = use_context_logger_level_on_not_set
        self._ignore_loggers_equal = ignore_loggers_equal
        self._ignore_loggers_like = ignore_loggers_like
        self._handle_origin_logger = handle_origin_logger

    def run(self):
        with self._action_logger.sub_logger(name=self._name,
                                            title=self._title,
                                            log_level=self._log_level,
                                            parent=self._parent,
                                            formatter=self._formatter,
                                            steps=self._steps,
                                            on_success_msg=self._on_success_msg,
                                            on_error_msg=self._on_error_msg,
                                            end_steps=self._end_steps,
                                            show_errors=self._show_errors,
                                            halt_on_error=self._halt_on_error,
                                            use_context_logger_level=self._use_context_logger_level,
                                            use_context_logger_level_on_not_set=self._use_context_logger_level_on_not_set,
                                            ignore_loggers_equal=self._ignore_loggers_equal,
                                            ignore_loggers_like=self._ignore_loggers_like,
                                            handle_origin_logger=self._handle_origin_logger) as sub_logger:
            self._sub_logger = sub_logger
            result = bool(self.payload())
            self._sub_logger.finalize(success=result)

    async def wait(self, timeout: int = -1):
        start_wait = time.perf_counter()
        while self.is_alive():
            if timeout != -1:
                if time.perf_counter() - start_wait > timeout:
                    self._sub_logger.finalize(success=False, on_error_msg="Thread timed out.")
            await asyncio.sleep(0.1)

    def payload(self) -> bool:
        if self._payload is None:
            raise NotImplementedError("Payload not implemented.")
        return self._payload(*self._payload_args, **self._payload_kwargs)



import string


def split(s: str, word_chars: str) -> list[str]:
    word_split = []
    word = ""
    for c in s:
        if c not in word_chars:
            word_split.append(word)
            word = ""
            continue
        word += c
    if len(word) > 0:
        word_split.append(word)
    return word_split


def lowercase(s: str) -> str:
    """
    Convert string to lowercase.
    Example: "Te-st test" -> "te-st test"

    :param s: Input string
    :return: String in lowercase
    """

    return s.lower()


def uppercase(s: str) -> str:
    """
    Convert string to uppercase.
    Example: "Te-st test" -> "TE-ST TEST"

    :param s: Input string
    :return: String in uppercase
    """

    return s.upper()


def camelcase(s: str) -> str:
    """
    Convert string to camelcase.
    Example: "Te-st test" -> "TeStTest"

    :param s: Input string
    :return: String in camelcase
    """

    case = "".join([word.capitalize() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])

    return case[0].lower() + case[1:]


def capitalcase(s: str) -> str:
    """
    Convert string to capitalcase.
    Example: "Te-st test" -> "Te St Test"

    :param s: Input string
    :return: String in capitalcase
    """

    return " ".join([word.capitalize() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def constantcase(s: str) -> str:
    """
    Convert string to constantcase.
    Example: "Te-st test" -> "TE_ST_TEST"

    :param s: Input string
    :return: String in constantcase
    """

    return "_".join([word.upper() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def dotcase(s: str) -> str:
    """
    Convert string to dotcase.
    Example: "Te-st test" -> "te.st.test"

    :param s: Input string
    :return: String in dotcase
    """

    return ".".join([word.lower() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def headercase(s: str) -> str:
    """
    Convert string to headercase.
    Example: "Te-st test" -> "Te-St-Test"

    :param s: Input string
    :return: String in headercase
    """

    return "-".join([word.capitalize() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def nocase(s: str) -> str:
    """
    Convert string to nocase.
    Example: "Te-st test" -> "te st test"

    :param s: Input string
    :return: String in nocase
    """

    return " ".join([word.lower() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def paramcase(s: str) -> str:
    """
    Convert string to paramcase.
    Example: "Te-st test" -> "te-st-test"

    :param s: Input string
    :return: String in paramcase
    """

    return "-".join([word.lower() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def pascalcase(s: str) -> str:
    """
    Convert string to pascalcase.
    Example: "Te-st test" -> "TeStTest"

    :param s: Input string
    :return: String in pascalcase
    """

    return "".join([word.capitalize() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


def pathcase(s: str) -> str:
    """
    Convert string to pathcase.
    Example: "Te-st test" -> "te/st/test"

    :param s: Input string
    :return: String in pathcase
    """

    return "/".join([word.lower() for word in split(s=s, word_chars=string.ascii_letters + string.digits)])


if __name__ == '__main__':
    c = pathcase("te-st test")
    print(c)



from datetime import datetime, timedelta
from typing import Optional

from mongoengine import Document, ReferenceField, DateTimeField, StringField, CASCADE
from starlette.requests import Request

from wiederverwendbar.starlette_admin.mongoengine.auth.settings import MongoengineAdminAuthSettings
from wiederverwendbar.starlette_admin.mongoengine.auth.documents.user import User


class Session(Document):
    meta = {"collection": "session"}

    user: User = ReferenceField(User, required=True, reverse_delete_rule=CASCADE)
    app_name: str = StringField(regex=r"^[a-zA-Z0-9_-]+$", required=True)
    user_agent: str = StringField(default="")
    created: datetime = DateTimeField(default=datetime.now, required=True)
    last_access: datetime = DateTimeField()

    async def __admin_repr__(self, request: Request):
        return f"{await self.user.__admin_repr__(request)} - {self.id})"

    @classmethod
    def get_session_from_request(cls, request: Request) -> Optional["Session"]:
        # get settings
        settings = MongoengineAdminAuthSettings.from_state(state=request.state)

        # expire sessions
        now = datetime.now()
        for session in cls.objects(app_name=settings.admin_name):
            expired = False
            if settings.admin_session_max_age is not None:
                if session.last_access + timedelta(seconds=settings.admin_session_max_age) < now:
                    expired = True
            if settings.admin_session_absolute_max_age is not None:
                if session.created + timedelta(seconds=settings.admin_session_absolute_max_age) < now:
                    expired = True
            if expired:
                session.delete()

        # get session id from session
        session_id = request.session.get("session_id", None)
        if session_id is None:
            return None

        # get session from database
        session = cls.objects(id=session_id, app_name=settings.admin_name).first()
        if session is None:
            return None

        return session



import logging

from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.datastructures import State
from starlette_admin.base import BaseAdmin as _BaseAdmin

logger = logging.getLogger(__name__)



class BaseAdmin(_BaseAdmin):
    def mount_to(self, app: Starlette) -> None:
        super().mount_to(app)

        # get admin app
        admin_app = None
        for app in app.routes:
            if not isinstance(app, Mount):
                continue
            if app.name != self.route_name:
                continue
            admin_app = app.app
        if admin_app is None:
            raise ValueError("Admin app not found")

        # add admin to admin app
        admin_app.state.admin = self

    @classmethod
    def from_state(cls, state: State):
        admin = state.admin
        if not isinstance(admin, cls):
            raise ValueError(f"Admin is not instance of {cls}")
        return admin



from wiederverwendbar.logger.handlers import (StreamConsoleHandler,
                                              RichConsoleHandler,
                                              TarRotatingFileHandler)
from wiederverwendbar.logger.context import LoggingContext
from wiederverwendbar.logger.file_modes import (FileModes)
from wiederverwendbar.logger.helper import (logger_exists,
                                            remove_logger)
from wiederverwendbar.logger.log_levels import (LogLevels)
from wiederverwendbar.logger.logger import (Logger)
from wiederverwendbar.logger.settings import (LoggerSettings)
from wiederverwendbar.logger.singleton import (LoggerSingleton)
from wiederverwendbar.logger.terminal_out_files import (TerminalOutFiles)



from typing import Optional, List, Union, Type

from starlette_admin.views import BaseView

from wiederverwendbar.starlette_admin.drop_down_icon_view.views import DropDownIconView


class AuthView(DropDownIconView):
    def __init__(self,
                 label: Optional[str] = None,
                 views: Optional[List[Union[Type[BaseView], BaseView]]] = None,
                 icon: Optional[str] = None,
                 always_open: Optional[bool] = None,
                 show_icon_on_sub_views: Optional[bool] = None
                 ):
        # set default values
        label = label or "Authentifizierung"
        views = views or []
        icon = icon or "fa fa-lock"
        if always_open is None:
            always_open = False
        if show_icon_on_sub_views is None:
            show_icon_on_sub_views = True

        super().__init__(label=label, views=views, icon=icon, always_open=always_open, show_icon_on_sub_views=show_icon_on_sub_views)



from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field.admin import BooleanAlsoAdmin
from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field.converter import BooleanAlsoConverter
from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field.field import BooleanAlsoField


import time
from typing import Union

from wiederverwendbar.threading import ExtendedThread



def test_watchdog(self: ExtendedThread):
    return True


def thread_target(*args, **kwargs):
    print("loop", args, kwargs)


if __name__ == '__main__':
    thread = ExtendedThread(name="test", target=thread_target, args=(1, 2), kwargs={"a": 1}, loop_sleep_time=1, watchdog_target=test_watchdog)

    while True:
        try:
            time.sleep(3)
            with thread.loop_wait():
                print("main thread")
                time.sleep(5)
        except KeyboardInterrupt:
            break
    # thread.stop()
    # thread.raise_exception(Exception("hallo welt"))



from datetime import datetime

from mongoengine import Document, DateTimeField, StringField, ListField, DictField
from mongoengine.base import TopLevelDocumentMetaclass


class MongoengineLogDocumentMeta(TopLevelDocumentMetaclass):
    ...


class MongoengineLogDocument(Document, metaclass=MongoengineLogDocumentMeta):
    meta = {"collection": "log",
            "allow_inheritance": True,
            'index_cls': False,
            "indexes": [
                {
                    "fields": ["timestamp"],
                    "expireAfterSeconds": 10 * 60 * 24 * 30  # 30 days
                }
            ]}

    timestamp: datetime = DateTimeField(required=True)
    entries: list[dict] = ListField(DictField(), required=True)



import logging
import asyncio
import threading
import time
from typing import Optional

import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from starlette.requests import Request
from starlette_admin.contrib.mongoengine import Admin, ModelView
from starlette_admin.actions import action
from mongoengine import Document, StringField
from kombu import Connection

from wiederverwendbar.mongoengine import MongoengineDbSingleton
from wiederverwendbar.starlette_admin import ActionLogAdmin, ActionLogger

logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)
logger.setLevel(logging.DEBUG)

# connect to database
MongoengineDbSingleton(init=True)

# create kombu connection
kombu_connection = Connection(MongoengineDbSingleton().connection_string)

# Create starlette app
app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse("<a href=/admin/>Click me to get to Admin!</a>"),
        ),
    ],
)


class MyAdmin(Admin, ActionLogAdmin):
    ...


# Create admin
admin = MyAdmin(title="Test Admin", kombu_connection=kombu_connection)


class Test(Document):
    meta = {"collection": "test"}

    test_str = StringField()


class TestView(ModelView):
    def __init__(self):
        super().__init__(document=Test, icon="fa fa-server", name="Test", label="Test")

    actions = ["delete", "test_action_normal", "test_action_action_log"]

    @action(name="test_action_normal",
            text="Test Action - Normal")
    # confirmation="Möchtest du die Test Aktion durchführen?",
    # icon_class="fa-regular fa-network-wired",
    # submit_btn_text="Ja, fortsetzen",
    # submit_btn_class="btn-success")
    async def test_action_normal(self, request: Request, pk: list[str]) -> str:
        await asyncio.sleep(2)

        return "Test Aktion erfolgreich."

    @action(name="test_action_action_log",
            text="Test Action - Action Log")
    # confirmation="Möchtest du die Test Aktion durchführen?",
    # icon_class="fa-regular fa-network-wired",
    # submit_btn_text="Ja, fortsetzen",
    # submit_btn_class="btn-success")
    async def test_action_action_log(self, request: Request, pk: list[str]) -> str:
        with ActionLogger(request, parent=logger) as action_logger:
            action_thread = ActionThread(action_logger=action_logger, payload=payload, name="action_thrad_1", title="Action Thread 1")
            action_thread.start()
            await action_thread.wait(timeout=5)

        return "Test Aktion erfolgreich."


class ActionThread(threading.Thread):
    def __init__(self,
                 action_logger: ActionLogger,
                 name: str,
                 payload: Optional[callable] = None,
                 payload_args: Optional[list] = None,
                 payload_kwargs: Optional[dict] = None,
                 title: Optional[str] = None,
                 log_level: int = logging.NOTSET,
                 parent: Optional[logging.Logger] = None,
                 formatter: Optional[logging.Formatter] = None,
                 steps: Optional[int] = None,
                 on_success_msg: Optional[str] = None,
                 on_error_msg: Optional[str] = "Something went wrong.",
                 show_errors: Optional[bool] = None,
                 halt_on_error: Optional[bool] = None,
                 use_context_logger_level: bool = True,
                 use_context_logger_level_on_not_set: Optional[bool] = None,
                 ignore_loggers_equal: Optional[list[str]] = None,
                 ignore_loggers_like: Optional[list[str]] = None,
                 handle_origin_logger: bool = True,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)

        self.lock = threading.Lock()
        self._payload = payload
        self._payload_args = payload_args if payload_args is not None else []
        self._payload_kwargs = payload_kwargs if payload_kwargs is not None else {}

        self._action_logger = action_logger
        self._sub_logger = None

        self._name = name
        self._title = title
        self._log_level = log_level
        self._parent = parent
        self._formatter = formatter
        self._steps = steps
        self._on_success_msg = on_success_msg
        self._on_error_msg = on_error_msg
        self._show_errors = show_errors
        self._halt_on_error = halt_on_error
        self._use_context_logger_level = use_context_logger_level
        self._use_context_logger_level_on_not_set = use_context_logger_level_on_not_set
        self._ignore_loggers_equal = ignore_loggers_equal
        self._ignore_loggers_like = ignore_loggers_like
        self._handle_origin_logger = handle_origin_logger

    def run(self):
        with self._action_logger.sub_logger(name=self._name,
                                            title=self._title,
                                            log_level=self._log_level,
                                            parent=self._parent,
                                            formatter=self._formatter,
                                            steps=self._steps,
                                            on_success_msg=self._on_success_msg,
                                            on_error_msg=self._on_error_msg,
                                            show_errors=self._show_errors,
                                            halt_on_error=self._halt_on_error,
                                            use_context_logger_level=self._use_context_logger_level,
                                            use_context_logger_level_on_not_set=self._use_context_logger_level_on_not_set,
                                            ignore_loggers_equal=self._ignore_loggers_equal,
                                            ignore_loggers_like=self._ignore_loggers_like,
                                            handle_origin_logger=self._handle_origin_logger) as sub_logger:
            self._sub_logger = sub_logger
            try:
                self.payload()
            except Exception as e:
                with self.lock:
                    self._sub_logger.finalize(success=False, on_error_msg=str(e))
            self._sub_logger.finalize(success=True)

    async def wait(self, timeout: int = -1):
        start_wait = time.perf_counter()
        while self.is_alive():
            if timeout != -1:
                if time.perf_counter() - start_wait > timeout:
                    with self.lock:
                        self._sub_logger.finalize(success=False, on_error_msg="Thread timed out.")
            await asyncio.sleep(0.1)

    def payload(self):
        if self._payload is None:
            raise NotImplementedError("Payload not implemented.")
        self._payload(*self._payload_args, **self._payload_kwargs)


def payload():
    while True:
        t_logger = logging.getLogger("test_foo")
        t_logger.debug("foo")
        print("foo")
        time.sleep(1)


# Add views to admin#
admin.add_view(TestView())

# Mount admin to app
admin.mount_to(app)

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)



from pathlib import Path

from wiederverwendbar.functions.test_file import test_file

if __name__ == '__main__':
    generated_file = test_file(1, 'MB')

    # move the file to the current directory
    generated_file = generated_file.replace(Path(generated_file.name))

    input("Press any key to remove the generated file.")

    # remove the file
    generated_file.unlink()



import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route

from starlette_admin.contrib.mongoengine.admin import Admin
from starlette_admin.contrib.mongoengine.view import ModelView

from wiederverwendbar.examples.mongoengine.automatic_reference import Test1, Test2
from wiederverwendbar.mongoengine import MongoengineDbSingleton

# connect to database
MongoengineDbSingleton(init=True)

# Create starlette app
app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse('<a href="/admin/">Click me to get to Admin!</a>'),
        ),
    ],
)

# Create admin
admin = Admin(title="Test Admin")



class Test1View(ModelView):
    def __init__(self):
        super().__init__(document=Test1, icon="fa fa-server", name="Test1", label="Test1")

class Test2View(ModelView):
    def __init__(self):
        super().__init__(document=Test2, icon="fa fa-server", name="Test2", label="Test2")


# Add views to admin
admin.add_view(Test1View())
admin.add_view(Test2View())

# Mount admin to app
admin.mount_to(app)

if __name__ == '__main__':
    uvicorn.run(app, host="0.0.0.0", port=8000)



from typing import Optional, Sequence

from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette_admin.i18n import I18nConfig
from starlette_admin.views import CustomView
from starlette_admin.auth import BaseAuthProvider

from wiederverwendbar.starlette_admin.settings.admin import SettingsAdmin
from wiederverwendbar.starlette_admin.form_max_fields.settings import FormMaxFieldsAdminSettings


class FormMaxFieldsAdmin(SettingsAdmin):
    settings_class = FormMaxFieldsAdminSettings

    def __init__(
            self,
            title: Optional[str] = None,
            base_url: Optional[str] = None,
            route_name: Optional[str] = None,
            logo_url: Optional[str] = None,
            login_logo_url: Optional[str] = None,
            templates_dir: Optional[str] = None,
            statics_dir: Optional[str] = None,
            index_view: Optional[CustomView] = None,
            auth_provider: Optional[BaseAuthProvider] = None,
            middlewares: Optional[Sequence[Middleware]] = None,
            session_middleware: Optional[type[SessionMiddleware]] = None,
            debug: Optional[bool] = None,
            i18n_config: Optional[I18nConfig] = None,
            favicon_url: Optional[str] = None,
            form_max_fields: Optional[int] = None,
            settings: Optional[FormMaxFieldsAdminSettings] = None
    ):
        super().__init__(
            title=title,
            base_url=base_url,
            route_name=route_name,
            logo_url=logo_url,
            login_logo_url=login_logo_url,
            templates_dir=templates_dir,
            statics_dir=statics_dir,
            index_view=index_view,
            auth_provider=auth_provider,
            middlewares=middlewares,
            session_middleware=session_middleware,
            debug=debug,
            i18n_config=i18n_config,
            favicon_url=favicon_url,
            settings=settings,
        )

        self.form_max_fields = form_max_fields or settings.form_max_fields

    async def _render_create(self, request: Request) -> Response:
        self._form_func = request.form
        request.form = self.form

        return await super()._render_create(request)

    async def _render_edit(self, request: Request) -> Response:
        self._form_func = request.form
        request.form = self.form

        return await super()._render_edit(request)

    async def form(self, *args, **kwargs):
        if "max_fields" not in kwargs:
            kwargs["max_fields"] = self.form_max_fields

        return await self._form_func(*args, **kwargs)



from pydantic import BaseModel, Field
from wiederverwendbar.functions.security.hashed_password import HashedPassword


class HashedPasswordModel(BaseModel, HashedPassword):
    encoding: str = Field(..., title="Encoding", description="The encoding of the hashed password.")
    hash_function: str = Field(..., title="Hash Function", description="The hash function of the hashed password.")
    interactions: int = Field(..., title="Interactions", description="The interactions of the hashed password.")
    key_length: int = Field(..., title="Key Length", description="The key length of the hashed password.")
    salt: str = Field(..., title="Salt", description="The salt of the hashed password.")
    hashed_password: str = Field(..., title="Hashed Password", description="The hashed password.")



import time

from wiederverwendbar.examples.mongoengine.logger import MyMongoengineLogDocument
from wiederverwendbar.mongoengine import MongoengineLogStreamer

if __name__ == '__main__':
    with MongoengineLogStreamer(MyMongoengineLogDocument) as streamer:
        while True:
            time.sleep(1)



import logging


class MongoengineLogFormatter(logging.Formatter):
    DEFAULT_PROPERTIES = logging.LogRecord('', 0, '', 0, '', (), (None, None, None), '').__dict__.keys()

    def format(self, record) -> dict:
        """
        Formats LogRecord into python dictionary.

        :param record: LogRecord instance.
        :return: dict
        """

        # standard entry
        entry = {
            'timestamp': record.created,
            'level': record.levelname,
            'thread': record.thread,
            'thread_name': record.threadName,
            'message': record.getMessage(),
            'logger_name': record.name,
            'file_name': record.pathname,
            'module': record.module,
            'method': record.funcName,
            'line_number': record.lineno
        }

        # add exception information if present
        if record.exc_info is not None:
            entry.update({
                'exception': {
                    'message': str(record.exc_info[1]),
                    'code': 0,
                    'stack_trace': self.formatException(record.exc_info)
                }
            })

        # add extra information
        if len(self.DEFAULT_PROPERTIES) != len(record.__dict__):
            contextual_extra = set(record.__dict__).difference(
                set(self.DEFAULT_PROPERTIES))
            if contextual_extra:
                for key in contextual_extra:
                    entry[key] = record.__dict__[key]

        return entry



from datetime import datetime, timezone, timedelta


def offset() -> int:
    """
    Calculate the offset between the local time and the UTC time in hours.

    :return: The offset between the local time and the UTC time in hours.
    """

    utc_now = datetime.now(timezone.utc)
    naive_utc_now = datetime(utc_now.year, utc_now.month, utc_now.day, utc_now.hour, utc_now.minute, utc_now.second, utc_now.microsecond)

    delta_seconds = datetime.now().timestamp() - naive_utc_now.timestamp()
    delta_hours = round(delta_seconds // 3600)

    return delta_hours


def local_now() -> datetime:
    """
    Get the current local time.

    :return: The current local time.
    """

    return datetime.now(tz=timezone(timedelta(hours=offset())))


def to_local(utc_time: datetime) -> datetime:
    """
    Convert a UTC time to a local time.

    :param utc_time: The UTC time to convert.
    :return: The local time.
    """

    return datetime.fromtimestamp(utc_time.timestamp() + offset() * 3600).replace(tzinfo=timezone(timedelta(hours=offset())))



from pathlib import Path
from typing import Optional

from pydantic import Field

from wiederverwendbar.starlette_admin.settings.settings import AdminSettings


class MongoengineAdminAuthSettings(AdminSettings):
    admin_auth: bool = Field(default=True, title="Admin Auth", description="Whether the admin panel requires authentication.")
    admin_login_path: str = Field(default="/login", title="Admin Login Path", description="The path of the login page of the admin panel.")
    admin_logout_path: str = Field(default="/logout", title="Admin Logout Path", description="The path of the logout page of the admin panel.")
    admin_allow_routes: Optional[list[str]] = Field(default=None, title="Admin Allow Routes", description="The routes that are allowed without authentication in the admin panel.")
    admin_user_app_title: str = Field(default="Hallo, {{session.user.username}}!", title="Admin User App Title", description="The title of the admin panel for the user. "
                                                                                                                             "If will be evaluated with the session.")
    admin_company_logo_dir: str = Field(default="company_logo", title="Admin Company Logo Subdirectory",
                                        description="The subdirectory of the admin panel static for the company logos.")
    admin_company_logos_suffixes: list[str] = Field(default=[".png", ".jpg", ".jpeg", ".gif"], title="Admin Company Logos Suffixes",
                                                    description="The suffixes of the company logos in the admin panel static.")
    admin_superuser_username: Optional[str] = Field(default="admin", title="Admin User Superuser username", description="The username of the superuser of the admin panel.")
    admin_superuser_auto_create: bool = Field(default=False, title="Admin User Superuser Auto Create",
                                              description="Whether the superuser is automatically created if it does not exist. The password will be the same as the username.")
    admin_user_allows_empty_password_login: bool = Field(default=False, title="Admin User Allows Empty Password Login",
                                                         description="Whether the user is allowed to login with an empty password.")

    @property
    def admin_static_company_logo_dir(self) -> Path:
        if self.admin_static_dir is None:
            raise FileNotFoundError("Admin static directory is not set. Please set it to the directory of the admin panel static.")

        company_logo_dir = Path(self.admin_static_dir) / self.admin_company_logo_dir
        if not company_logo_dir.is_dir():
            raise FileNotFoundError(f"Admin company logo directory {company_logo_dir} is not a directory.")
        return company_logo_dir



from wiederverwendbar.starlette_admin.mongoengine.auth import (AuthView,
                                                               Session,
                                                               SessionView,
                                                               User,
                                                               UserView,
                                                               MongoengineAuthAdmin,
                                                               MongoengineAdminAuthProvider,
                                                               MongoengineAdminAuthSettings)
from wiederverwendbar.starlette_admin.mongoengine.boolean_also_field import (BooleanAlsoAdmin,
                                                                             BooleanAlsoConverter,
                                                                             BooleanAlsoField)
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field import (GenericEmbeddedAdmin,
                                                                                          GenericEmbeddedConverter,
                                                                                          GenericEmbeddedDocumentField,
                                                                                          ListField,
                                                                                          GenericEmbeddedDocumentView)
from wiederverwendbar.starlette_admin.mongoengine.ipv4_field import (IPv4Converter)
from wiederverwendbar.starlette_admin.mongoengine.converter import (MongoengineConverter)
from wiederverwendbar.starlette_admin.mongoengine.view import (FixedModelView, MongoengineModelView)



from typing import Optional, Sequence

from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette_admin.i18n import lazy_gettext as _
from starlette_admin.i18n import I18nConfig
from starlette_admin.views import CustomView
from starlette_admin.auth import BaseAuthProvider

from wiederverwendbar.starlette_admin.admin import BaseAdmin
from wiederverwendbar.starlette_admin.settings.settings import AdminSettings


class SettingsAdminMeta(type):
    def __new__(cls, name, bases, attrs):
        # combine settings_cls from bases and attrs
        all_settings_classes = []
        # get settings_class from attrs
        if "settings_class" in attrs:
            # skip duplicates
            if attrs["settings_class"] not in all_settings_classes:
                # add at the beginning of the combined list
                all_settings_classes.append(attrs["settings_class"])

        # get all settings_classes from bases
        for base in bases:
            if hasattr(base, "settings_classes"):
                for settings_class in base.settings_classes:
                    # skip duplicates
                    if settings_class in all_settings_classes:
                        continue
                    # add at the beginning of the combined list
                    all_settings_classes.append(settings_class)

        # set settings_classes to the combined list
        attrs["settings_classes"] = all_settings_classes

        return super().__new__(cls, name, bases, attrs)


class SettingsAdmin(BaseAdmin, metaclass=SettingsAdminMeta):
    settings_class = AdminSettings
    settings_classes: Sequence[type[AdminSettings]] = []

    def __init__(
            self,
            title: Optional[str] = None,
            base_url: Optional[str] = None,
            route_name: Optional[str] = None,
            logo_url: Optional[str] = None,
            login_logo_url: Optional[str] = None,
            templates_dir: Optional[str] = None,
            statics_dir: Optional[str] = None,
            index_view: Optional[CustomView] = None,
            auth_provider: Optional[BaseAuthProvider] = None,
            middlewares: Optional[Sequence[Middleware]] = None,
            session_middleware: Optional[type[SessionMiddleware]] = None,
            debug: Optional[bool] = None,
            i18n_config: Optional[I18nConfig] = None,
            favicon_url: Optional[str] = None,
            settings: Optional[AdminSettings] = None
    ):
        # get settings from the settings class if not provided
        settings = settings or self.settings_class()
        for settings_class in self.settings_classes:
            if not isinstance(settings, settings_class):
                raise ValueError(f"settings must be an instance of {settings_class.__name__}")

        # set the values from the settings class if not provided
        title = _(title) or _(settings.admin_title)
        base_url = base_url or settings.admin_base_url
        route_name = route_name or settings.admin_route_name
        logo_url = logo_url or settings.admin_logo_url
        login_logo_url = login_logo_url or settings.admin_login_logo_url
        templates_dir = templates_dir or settings.admin_templates_dir
        statics_dir = statics_dir or settings.admin_static_dir
        auth_provider = auth_provider or None
        # set middlewares
        middlewares = middlewares or []

        # set session middleware
        session_middleware = session_middleware or SessionMiddleware
        if not any(issubclass(middleware.cls, SessionMiddleware) for middleware in middlewares):
            middlewares.append(Middleware(session_middleware,  # noqa
                                          secret_key=settings.admin_session_secret_key,
                                          session_cookie=settings.admin_session_cookie,
                                          max_age=settings.admin_session_max_age,
                                          path=settings.admin_session_path,
                                          same_site=settings.admin_session_same_site.value,
                                          https_only=settings.admin_session_https_only,
                                          domain=settings.admin_session_domain))

        debug = debug if debug is not None else settings.admin_debug
        i18n_config = i18n_config or I18nConfig(default_locale=settings.admin_language.value,
                                                language_cookie_name=settings.admin_language_cookie_name,
                                                language_header_name=settings.admin_language_header_name,
                                                language_switcher=settings.admin_language_available)
        favicon_url = favicon_url or settings.admin_favicon_url

        super().__init__(
            title=title,
            base_url=base_url,
            route_name=route_name,
            logo_url=logo_url,
            login_logo_url=login_logo_url,
            templates_dir=templates_dir,
            statics_dir=statics_dir,
            index_view=index_view,
            auth_provider=auth_provider,
            middlewares=middlewares,
            debug=debug,
            i18n_config=i18n_config,
            favicon_url=favicon_url,
        )

        # set the settings
        self.settings = settings



import os

from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route

from wiederverwendbar.logger import LoggerSingleton, LoggerSettings
from wiederverwendbar.singleton import Singleton
from wiederverwendbar.uvicorn import UvicornServer

init = False
try:
    Singleton.get_by_type(LoggerSingleton)
    init = True
except RuntimeError:
    ...

if not init or __name__ == '__main__':
    LoggerSingleton(name="test", settings=LoggerSettings(log_level=LoggerSettings.LogLevels.DEBUG), init=True)

app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse('<h1>Hello, world!</h1>'),
        ),
    ],
)

if __name__ == '__main__':
    UvicornServer("uvicorn_server:app", host="0.0.0.0", port=8000, workers=1)
    print()



import asyncio
import logging
from pathlib import Path
from warnings import warn
from threading import Thread, Lock
from typing import Union, Any
from socket import timeout as socket_timeout

import nest_asyncio
from jinja2 import PackageLoader
from pydantic import ValidationError
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import FileResponse
from starlette.routing import WebSocketRoute, Route
from starlette.endpoints import WebSocketEndpoint
from starlette.websockets import WebSocket, WebSocketState
from starlette.types import Scope, Receive, Send
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette_admin.i18n import I18nConfig
from starlette_admin.views import CustomView
from starlette_admin.auth import BaseAuthProvider
from kombu import Connection, Exchange, Queue, Producer

from typing import Optional, Sequence

from wiederverwendbar.starlette_admin.action_log.settings import ActionLogAdminSettings
from wiederverwendbar.starlette_admin.settings.admin import SettingsAdminMeta, SettingsAdmin
from wiederverwendbar.starlette_admin.multi_path.admin import MultiPathAdminMeta, MultiPathAdmin
from wiederverwendbar.starlette_admin.action_log.logger import ActionLogger, ActionLoggerResponse

logger = logging.getLogger(__name__)


class ActionLogAdminMeta(SettingsAdminMeta, MultiPathAdminMeta):
    ...


class ActionLogAdmin(SettingsAdmin, MultiPathAdmin, metaclass=ActionLogAdminMeta):
    settings_class = ActionLogAdminSettings
    static_files_packages = [("wiederverwendbar", "starlette_admin/action_log/statics")]
    template_packages = [PackageLoader("wiederverwendbar", "starlette_admin/action_log/templates")]

    class ActionLogEndpoint(WebSocketEndpoint):
        encoding = "text"
        wait_for_logger_timeout = 5

        def __init__(self, scope: Scope, receive: Receive, send: Send):
            super().__init__(scope=scope, receive=receive, send=send)

            self.loop = asyncio.get_event_loop()
            self.websocket: Optional[WebSocket] = None
            self.connection: Optional[Connection] = None
            self.exchange: Optional[Exchange] = None
            self.start_queue: Optional[Queue] = None
            self.log_queue: Optional[Queue] = None
            self.response_queue: Optional[Queue] = None
            self.download_queue: Optional[Queue] = None
            self.exit_queue: Optional[Queue] = None
            self.producer: Optional[Producer] = None
            self.log_thread = Thread(target=self.receive_logs)
            self.lock = Lock()

        @property
        def ready(self):
            with self.lock:
                if self.connection is None:
                    return False
                if self.exchange is None:
                    return False
                if self.start_queue is None:
                    return False
                if self.log_queue is None:
                    return False
                if self.response_queue is None:
                    return False
                if self.download_queue is None:
                    return False
                if self.exit_queue is None:
                    return False
                if self.producer is None:
                    return False
                if self.websocket is None:
                    return False
                if self.websocket.client_state != WebSocketState.CONNECTED:
                    return False
                return True

        def receive_logs(self):
            with self.connection.Consumer([self.log_queue], callbacks=[self.send_log]):
                while True:
                    if not self.ready:
                        break
                    try:
                        self.connection.drain_events(timeout=0.001)
                    except socket_timeout:
                        ...

        def send_log(self, body, message):
            if not self.ready:
                warn(f"Cannot send log to websocket, because it is not connected -> {body}", UserWarning)
            with self.lock:
                asyncio.run_coroutine_threadsafe(self.websocket.send_json(body), self.loop)
            message.ack()

        def response(self, data: Union[str, dict[str, Any], ActionLoggerResponse]) -> bool:
            if not isinstance(data, ActionLoggerResponse):
                try:
                    data = ActionLogger.parse_response_obj(data)
                except ValidationError:
                    return False
            data_dict = data.model_dump()

            self.producer.publish(data_dict, exchange=self.exchange, routing_key=self.response_queue.name)

            return True

        async def on_connect(self, websocket: WebSocket):
            if self.websocket is not None:
                await websocket.close(code=1008)
                return

            # get kombu connection
            self.connection = ActionLogger.get_kombu_connection(request_or_websocket=websocket)

            # create exchange and queues from websocket request
            self.exchange, self.start_queue, self.log_queue, self.response_queue, self.download_queue, self.exit_queue = ActionLogger.get_action_log_queues(
                request_or_websocket=websocket)

            # create producer
            self.producer = self.connection.Producer(serializer='json')

            # accept websocket
            await websocket.accept()

            # save websocket
            self.websocket = websocket

            # start log thread
            self.log_thread.start()

            # send start message to logger
            self.producer.publish({"start": "start"}, exchange=self.exchange, routing_key=self.start_queue.name)

        async def on_receive(self, websocket: WebSocket, data: str):
            # send response to logger
            if not self.response(data=data):
                # close websocket
                await websocket.close(code=1008)

        async def on_disconnect(self, websocket: WebSocket, close_code: int):
            self.producer.publish({"exit": "exit"}, exchange=self.exchange, routing_key=self.exit_queue.name)

            with self.lock:
                self.connection = None
                self.exchange = None
                self.start_queue = None
                self.log_queue = None
                self.response_queue = None
                self.download_queue = None
                self.exit_queue = None
                self.producer = None
                self.websocket = None

    def __init__(
            self,
            kombu_connection: Connection,
            title: Optional[str] = None,
            base_url: Optional[str] = None,
            route_name: Optional[str] = None,
            logo_url: Optional[str] = None,
            login_logo_url: Optional[str] = None,
            templates_dir: Optional[str] = None,
            statics_dir: Optional[str] = None,
            index_view: Optional[CustomView] = None,
            auth_provider: Optional[BaseAuthProvider] = None,
            middlewares: Optional[Sequence[Middleware]] = None,
            session_middleware: Optional[type[SessionMiddleware]] = None,
            debug: Optional[bool] = None,
            i18n_config: Optional[I18nConfig] = None,
            favicon_url: Optional[str] = None,
            settings: Optional[ActionLogAdminSettings] = None
    ):
        super().__init__(
            title=title,
            base_url=base_url,
            route_name=route_name,
            logo_url=logo_url,
            login_logo_url=login_logo_url,
            templates_dir=templates_dir,
            statics_dir=statics_dir,
            index_view=index_view,
            auth_provider=auth_provider,
            middlewares=middlewares,
            session_middleware=session_middleware,
            debug=debug,
            i18n_config=i18n_config,
            favicon_url=favicon_url,
            settings=settings
        )

        self.kombu_connection = kombu_connection

    def init_routes(self) -> None:
        super().init_routes()
        self.routes.append(WebSocketRoute(path="/ws/action_log/{action_log_key}", endpoint=self.ActionLogEndpoint, name="action_log"))  # noqa
        self.routes.append(Route("/action_log/download/{action_log_key}", self.action_log_download, methods=["GET"], name="action_log_download"))  # noqa

        nest_asyncio.apply()  # ToDo: ugly hack to make asyncio.run work outside of debug mode, remove if it's not needed anymore

    def action_log_download(self, request: Request):
        # get action_log_key from request
        action_log_key = request.path_params.get("action_log_key", None)
        if action_log_key is None:
            raise HTTPException(status_code=404, detail="Action Log Key not provided.")

        # get download queue
        exchange, start_queue, log_queue, response_queue, download_queue, exit_queue = ActionLogger.get_action_log_queues(request_or_websocket=request)

        download_file: Optional[Path] = None

        def send_file(body, message):
            nonlocal download_file

            download_file = Path(body["file_path"])
            # message.ack()

        # get download file from download queue
        with self.kombu_connection.Consumer([download_queue], callbacks=[send_file]):
            try:
                self.kombu_connection.drain_events(timeout=5)
            except socket_timeout:
                raise HTTPException(status_code=500, detail="Timeout while waiting for download file.")

        if download_file is None:
            raise HTTPException(status_code=500, detail="No download file received.")
        elif not download_file.is_file():
            raise HTTPException(status_code=404, detail="Download file not found.")

        return FileResponse(download_file, filename=download_file.name)
